#ifndef GRAPHISO_H
#define GRAPHISO_H

#include "File.h"

/// Takes a template argument of any type 'T' and the nodes of type 'int', which are partitioned into 'partitionedNodes'.
template <typename T>
void PartitionNodes(const std::map<int, T>& invariant, Vector2D& partitionedNodes)
{
    std::map<T, std::vector<int>> nodeGroup;
    for(auto it = invariant.begin(); it != invariant.end(); ++it){
        auto it_g = nodeGroup.find(it->second);
        if(it_g == nodeGroup.end()){
            std::vector<int> tmp;
            tmp.push_back(it->first);
            nodeGroup.insert(std::make_pair(it->second, tmp));
        }
        else
            it_g->second.push_back(it->first);
    }
    /// Collect the partitions with decreasing vertex degrees.
    for(auto it = nodeGroup.rbegin(); it != nodeGroup.rend(); ++it)
        partitionedNodes.push_back(it->second);
}

class GraphIso
{
    public:
        GraphIso();
        virtual ~GraphIso();
        void GenerateCanonical(const GraphParameter& subgraph, std::vector<int>& canonicalForm, const IsoTestMap& dataGraph);
    protected:
    private:
        void ComputePermutationGroups(const GraphParameter& pattern, const IsoTestMap& dataGraph, Vector2D& permutationGroups);
        void ComputeVertexPermutation(Vector2D& permutationGroups, Vector2D& vertexPermutations);
};

#endif // GRAPHISO_H
