#ifndef SUBGRAPHMINER_H
#define SUBGRAPHMINER_H

#include "File.h"
#include "Sumgra.h"
#include "Frequency.h"
#include "Mis.h"
#include "GraphIso.h"
#include "TracesIso.h"
#include <../sumgra/Trie.h>


typedef std::map<int, std::set<std::vector<int>>> IsoMeasure;
//typedef std::map<int, std::vector<sparsegraph>> IsoMeasureNauty;


//struct TimeEval{
//    double SgExt = 0;
//    double SgFrq = 0;
//    double EmbEv = 0;
//};

class SubgraphMiner
{
    public:
        SubgraphMiner();
        virtual ~SubgraphMiner();
        void CollectSubgraphSeeds(const GraphParameter& dataGraphInfo, std::deque<Subgraph>& subgraphSeeds);
        void ExtractInfo(const Subgraph& subgraph, GraphParameter& subgraphInfo);
        void MineFrequentSubgraphs(const GraphParameter& dataGraphInfo, IndexType& graphIndexes, const int& supportVal, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg);
        void SubgraphExtension(const IsoTestMap& dataGraph, Subgraph& oldFSG, const std::set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& imeEval);
        void ExtendWithNode(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval);
        void ExtendWithoutNode(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval);
        void ExtendWithNodeNew(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval);
        void ExtendWithoutNodeNew(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval);
//        void SubgraphExtension(const IsoTestMap& dataGraph, Subgraph& oldFSG, const std::set<int>& newEdge, IsoMeasureNauty& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newFSGs);

        void FindFrequentSeeds(const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newSubgraphs, std::deque<Subgraph>& newFSGs);
////        void FindFrequentSubgraphs(TimeEval& timeEval, const std::set<int> newEdge, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, Subgraph& oldFSG, std::deque<Subgraph>& newSubgraphs, std::deque<Subgraph>& newFSGs);

        bool IsIsomorphicCurrent(Subgraph& newSubgraph, std::set<std::vector<int>>& extraNodeCanonical, const IsoTestMap& dataGraph);
        bool IsIsomorphicPrevious(const int& subgraphSize, const std::vector<int>& canonical, IsoMeasure& canonicalAllLevel);
//        bool IsNotIsomorphicCurrentNauty(Subgraph& newSubgraph, std::vector<sparsegraph>& extraNodeCanonical, const IsoTestMap& dataGraph);
//        bool IsNotIsomorphicPreviousNauty(const int& subgraphSize, sparsegraph& canonical, IsoMeasureNauty& canonicalAllLevel);
        void getInvalidCands(Subgraph& oldFSG, std::deque<Subgraph>& oldFSGs, FSGs& fsgs, Subgraph& newSubgraph, IndexType& newSubgraphInd);
        bool IsFrequent(const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, Subgraph& newSubgraph);
        bool InvalidTrieEdge(Trie* trie, std::set<int> newEdge);
        void DfsRecursiveMining(StackFSG& fsgClassStack, IsoMeasure& canonicalAllLevel, std::deque<Subgraph>& extendableEdges, int nE, TimeEval& timeEval, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg, std::deque<IndexType>& maximalFSG, FSGs& fsgs, int& tg);
//        void DfsRecursiveMining(StackFSG& fsgClassStack, IsoMeasureNauty& canonicalAllLevel, std::deque<Subgraph>& extendableEdges, int nE, TimeEval& timeEval, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg);
        void ArrangeByCoP(std::deque<Subgraph>& fsgSeedsSorted, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& basicNonFSG, std::deque<IndexType>& maximalFSG, FSGs& fsgs, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg, CoP& cop, TimeEval& timeEval);
        void CoPExtension(StackFSG& fsgClassStack, IsoMeasure& canonicalAllLevel, std::deque<Subgraph>& extendableEdges, int nE, TimeEval& timeEval, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg, std::deque<IndexType>& maximalFSG, FSGs& fsgs, int& tg);
    protected:
    private:

};

#endif // SUBGRAPHMINER_H
