/* This program demonstrates how an isomorphism is found between
   two graphs, using the Moebius graph as an example.
   This version uses Traces.
*/

#include "traces.h"

int main(int argc, char *argv[])
{
    DYNALLSTAT(int,lab1,lab1_sz);
    DYNALLSTAT(int,lab2,lab2_sz);
    DYNALLSTAT(int,ptn,ptn_sz);
    DYNALLSTAT(int,orbits,orbits_sz);
    DYNALLSTAT(int,map,map_sz);

    static DEFAULTOPTIONS_TRACES(options);
    TracesStats stats;

    int n,m,i;

 /* Select option for canonical labelling */

    options.getcanon = TRUE;

    n = 3;
    m = SETWORDSNEEDED(n);
    nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);

    DYNALLOC1(int,lab1,lab1_sz,n,"malloc");
    DYNALLOC1(int,lab2,lab2_sz,n,"malloc");
    DYNALLOC1(int,ptn,ptn_sz,n,"malloc");
    DYNALLOC1(int,orbits,orbits_sz,n,"malloc");
    DYNALLOC1(int,map,map_sz,n,"malloc");


/// Generate your own graphs
// A triangle graph
/*
    DYNALLSTAT(int,lab3,lab3_sz);
    SG_DECL(sg3); SG_DECL(cg3);
    SG_ALLOC(sg3,n,2*n,"malloc");
    sg3.nv = n;
    sg3.nde = 2*n;

    for (i = 0; i < n; ++i)
    {
      sg3.v[i] = 2*i;
      sg3.d[i] = 2;
      sg3.e[2*i] = (i+n-1)%n;
      sg3.e[2*i+1] = (i+n+1)%n;
    }
*/

// A wedge graph (enumeration-1)
    DYNALLSTAT(int,lab3,lab3_sz);
    /* Declare and initialize sparse graph structures */
    SG_DECL(sg3); SG_DECL(cg3);
    SG_ALLOC(sg3,n,4,"malloc");
    sg3.nv = n;
    sg3.nde = 4;

    for (i = 0; i < n; ++i)
    {
      sg3.v[i] = 2*i;
    }
    sg3.d[0] = 1; sg3.d[1] = 1; sg3.d[2] = 2;
    sg3.e[0] = 2; sg3.e[2] = 2; sg3.e[4] = 0; sg3.e[5] = 1;

// A wedge graph (enumeration-2)
    DYNALLSTAT(int,lab4,lab4_sz);
    SG_DECL(sg4); SG_DECL(cg4);
    SG_ALLOC(sg4,n,4,"malloc");
    sg4.nv = n;
    sg4.nde = 4;

    for (i = 0; i < n; ++i)
    {
      sg4.v[i] = 2*i;
    }
    sg4.d[0] = 2; sg4.d[1] = 1; sg4.d[2] = 1;
    sg4.e[0] = 1; sg4.e[1] = 2; sg4.e[3] = 0; sg4.e[5] = 0;

// Compute canonical representations;
    Traces(&sg3,lab1,ptn,orbits,&options,&stats,&cg3);
    Traces(&sg4,lab2,ptn,orbits,&options,&stats,&cg4);

 /* Compare canonically labelled graphs */

    if (aresame_sg(&cg3,&cg4))
    {
        printf("Isomorphic.\n");
        if (n <= 1000)
        {
         // Write the isomorphism.  For each i, vertex lab1[i] of sg1 maps onto vertex lab2[i] of sg2.  We compute the map in order of labelling because it looks better. //
            for (i = 0; i < n; ++i) map[lab1[i]] = lab2[i];
            for (i = 0; i < n; ++i) printf(" %d-%d",i,map[i]);
            printf("\n");
        }
    }
    else
        printf("Not isomorphic.\n");
    exit(0);
}
