#include "Frequency.h"

Frequency::Frequency()
{
    //ctor
}

Frequency::~Frequency()
{
    //dtor
}


bool EdgeExists(const std::vector<int>& p1, const std::vector<int>& p2)
{
    // Both p1 and p2 are set of vertices ordered as to maintain the adjacent connectivity; thus we can seamlessly retrieve the edges
    std::set<std::pair<int, int>> edges;
    for(size_t i = 0; i < p1.size()-1; ++i){
        edges.insert(std::make_pair(p1[i],p1[i+1]));
        edges.insert(std::make_pair(p1[i+1],p1[i]));
    }
    for(size_t i = 0; i < p2.size()-1; ++i){
        if(edges.find(std::make_pair(p2[i], p2[i+1])) != edges.end())
            return true; // two patterns share at least a common edge |
    }
}

void CreateEdgeOverlapGraph(const Vector2D& embeddings, Vector2D& adjList)
{
    /// Creates an overlap graph from the set of embeddings where an edge exists if two embeddings share an edge.

    /// Verify this approach when you use it.
    for(size_t i = 0; i < embeddings.size()-1; ++i){
        for(size_t j = i; j < embeddings.size(); ++j){
            if(EdgeExists(embeddings[i], embeddings[j])){
                adjList[i].push_back(j);
                adjList[j].push_back(i);
            }
        }
    }
}

void CreateNodeOverlapGraph(const Vector2D& embeddings, Vector2D& adjList)
{
    /// Creates an overlap graph from the set of embeddings where an edge exists if two embeddings share a vertex.
    int subgraphSize = embeddings[0].size();
    int nEmbeddings = embeddings.size();
    for(size_t i = 0; i < nEmbeddings-1; ++i){
        std::set<int> vertexSet(embeddings[i].begin(), embeddings[i].end());
        for(size_t j = i+1; j < nEmbeddings; ++j){
            bool overlapNotFound = true;
            size_t k = 0;
            while(overlapNotFound && k < subgraphSize){
                if(vertexSet.find(embeddings[j][k]) != vertexSet.end()){
                    adjList[i].push_back(j);
                    adjList[j].push_back(i);
                    overlapNotFound = false;
                }
                ++k;
            }
        }
    }
}

bool Frequency::IsMisFrequent(const Vector2D& embeddings, const int& freq)
{
    Vector2D adjList(embeddings.size());
    CreateNodeOverlapGraph(embeddings, adjList);
//    CreateEdgeOverlapGraph(embeddings, adjList);

    Mis mIS;
    int independenceNo; // independence number for MIS
    mIS.FindMis(adjList, freq, independenceNo); // adjList can contain disconnected nodes that have no neighbours. Check if 'FindMis' can handle this well |

//for(size_t i = 0; i < embeddings.size(); ++i){
//    for(size_t j = 0; j < embeddings[i].size(); ++j)
//        cout << embeddings[i][j] << ", ";
//    cout << endl;
//}
//for(size_t i = 0; i < adjList.size(); ++i){
//    cout << i << ": ";
//    for(size_t j = 0; j < adjList[i].size(); ++j)
//        cout << adjList[i][j] << ", ";
//    cout << endl;
//}
//cout << endl;
//cout << "emb size: " << embeddings.size() << endl;
//cout << "MIS size: " << independenceNo << endl;
    if(independenceNo >= freq)
        return true;
    else
        return false;
}


bool Frequency::IsMinImageFrequent(const Vector2D& embeddings, const int& freq)
{

    bool MinReached = true;
    int subgraphSize = embeddings[0].size();
    for(size_t i = 0; i < subgraphSize; ++i){
        std::unordered_set<int> distinctNodeMatches;
        int j = 0;
        while(distinctNodeMatches.size() < freq && j < embeddings.size()){
            distinctNodeMatches.insert(embeddings[j][i]);
            ++j;
        }
        if(distinctNodeMatches.size() < freq){
            MinReached = false;
            break;
        }
    }
    return MinReached;


/*
    int minImage = embeddings.size(); // Set the image value to the maximal possible value for a given set of embeddings |
    int subgraphSize = embeddings[0].size();
    for(size_t i = 0; i < subgraphSize; ++i){
        std::unordered_set<int> distinctNodeMatches;
        for(size_t j = 0; j < embeddings.size(); ++j)
            distinctNodeMatches.insert(embeddings[j][i]);
        if(distinctNodeMatches.size() < minImage)
            minImage = distinctNodeMatches.size();
    }
    if(minImage >= freq){
        return true;
    }
    else
        return false;
*/
}
