#include "SubgraphMiner.h"

SubgraphMiner::SubgraphMiner()
{
    //ctor
}

SubgraphMiner::~SubgraphMiner()
{
    //dtor
}


bool TestIsoInvariants(const Subgraph& A, const Subgraph& B) // Can NOT be used for IsNotIsomorphicPrevious; but only for IsNotIsomorphicCurrent |
{
    /// Check node numbers
    if(A.parameters.nNodes != B.parameters.nNodes)
        return false;

    /// Check Node labels


    /// Check edges
    if(A.edgeLabels.size() != B.edgeLabels.size())
        return false;


    /// Check Edge Labels (multiedges).
    // Edge labels (multiedges)
    if(A.multiEdges.size() == B.multiEdges.size()){ // check the unique multiedges
        auto itB = B.multiEdges.begin();
        for(auto itA = A.multiEdges.begin(); itA != A.multiEdges.end(); ++itA){
            if(*itA != *itB)
                return false;
            ++itB;
        }
    }
    else
        return false;
}

void FetchInvalids(Subgraph& newSubgraph, Subgraph& oldSubgraph, IndexType& newSubgraphInd)
{
    Sumgra subIso;
    Vector2D solutions; // Holds a set of isomorphic solutions for "oldSubgraph.orderedNodes" |
    if(subIso.FindIsoMatch(oldSubgraph, newSubgraphInd, solutions)){
        for(size_t i = 0; i < solutions.size(); ++i){
            for(auto itO = oldSubgraph.orderedNodes.begin(), itN = solutions[i].begin(); itO != oldSubgraph.orderedNodes.end(); ++itO, ++itN){
    /// Make sure that the mapping of vertices with same vertex-ids is maintained between the old parent and new child. They have to be ISOMORPHIC NODES since we are combining invalid candidates for a node of a pattern.
                auto itIO = oldSubgraph.invalidCands.find(*itO);
                auto itIN = newSubgraph.invalidCands.find(*itN);
                if(itIN != newSubgraph.invalidCands.end() && itIO != oldSubgraph.invalidCands.end()){
                    if(itIO->second.size() > itIN->second.size()){
//cout << "BEFORE: " << itIN->second.size() << ",  " << itIO->second.size() << endl;
                        itIN->second = Union(itIO->second , itIN->second);
//cout << "AFTER:  " << itIN->second.size() << ",  " << itIO->second.size() << endl;
                    }
                }
            }
        }
    }
}

void SubgraphMiner::getInvalidCands(Subgraph& oldFSG, std::deque<Subgraph>& oldFSGs, FSGs& fsgs, Subgraph& newSubgraph, IndexType& newSubgraphInd)
{
/// Propagate the invalid candidates directly form a parent to its child.
    newSubgraph.invalidCands = oldFSG.invalidCands;


/// Propagate the invalid candidates from already discovered NFSG that are subgraphs of "newSubgraph". [subgraph checking is already done somewhere]. Try to add invalid candidates from there.

/// Propagate the invalid candidates from a parent along with its siblings to its child.
    /// MIght not be very effective as the siblings might not have collected any invalid candidates yet.


    if(newSubgraph.edgeLabels.size() > 2){ // oldFSG has to be at least size 2 |
        /// Propagate the invalid candidates from already discovered FSG that are subgraphs of "newSubgraph".

        for(auto it = fsgs.begin(); it != fsgs.end(); ++it){
            if(it->first >= newSubgraph.edgeLabels.size())
                continue;
            for(size_t i = 0; i < it->second.size(); ++i){
//                if(it->second[i].canonicalForm == oldFSGs.front().canonicalForm) // avoid repeated collection of invalid candidates |
//                    continue; // NOT NECESSARY; it slows down :(
            FetchInvalids(newSubgraph, it->second[i], newSubgraphInd);
            }
        }

        /// Propagate invalid candidates from the class of patterns.
        for(size_t i = 0; i < oldFSGs.size(); ++i)
            FetchInvalids(newSubgraph, oldFSGs[i], newSubgraphInd);

    }

//for(auto it = newSubgraph.invalidCands.begin(); it != newSubgraph.invalidCands.end(); ++it)
//    cout << it->first << ": " << it->second.size() << endl;

}

/// Decompose the set into all possible (2^n-1) subsets and update to the collection of all multiedges.
void RetrieveAllItems(std::set<int> inputSet, VecOfSet& allSubsets)
{
    /// Initilize the solution |
    std::set<int> init;
    auto it = inputSet.begin();
    init.insert(*it);
    allSubsets.push_back(init);

    if(inputSet.size() > 1){
        std::advance(it, 1);
        for(; it != inputSet.end(); ++it){
            /// Add the new current element as a new subset|
            std::set<int> single;
            single.insert(*it);
            allSubsets.push_back(single);
            /// Grow the already existing subsets with the new element |
            VecOfSet alreadyCollected = allSubsets;
            for(size_t k = 0; k < alreadyCollected.size(); ++k){
                alreadyCollected[k].insert(*it);
                allSubsets.push_back(alreadyCollected[k]);
            }
        }
    }
}

void CollectAllDataEdges(const std::map<std::set<int>, int>& distinctDataEdges, std::set<std::set<int>>& distinctEdges)
{
    for(auto it = distinctDataEdges.begin(); it != distinctDataEdges.end(); ++it){
        VecOfSet allSubsets;
        RetrieveAllItems(it->first, allSubsets);
        for(auto itS = allSubsets.begin(); itS != allSubsets.end(); ++itS)
            distinctEdges.insert(*itS);
    }
}



void SubgraphMiner::CollectSubgraphSeeds(const GraphParameter& dataGraphInfo, std::deque<Subgraph>& subgraphSeeds)
{
    if(dataGraphInfo.nodeLabelsExist){ // vertex labels are accounted for |
        cout << "code unwritten!" << endl;
    }
    else{ // no vertex labels |
        /// Collect all distinct multiedges in the datagraph and count their number of occurrences.
//        Trie * edgeTrie = new Trie(); // Maintain a multiedge trie structure |
        std::map<std::set<int>, int> distinctDataEdges; // <multiedge, its frequency>
        for(auto it = dataGraphInfo.eLabelMap.begin(); it != dataGraphInfo.eLabelMap.end(); ++it){
            auto itU = distinctDataEdges.find(it->second);
            if(itU == distinctDataEdges.end()){
                distinctDataEdges.insert(std::make_pair(it->second, 1));
//                edgeTrie->AddMultiedges(it->second); // add only when it is a new multiedge |
            }
            else
                ++itU->second;
        }
cout << "# distinct multiedge set: " << distinctDataEdges.size() << endl;

        /// NOT collecting the sub-multiedges.
/*
            /// Create structure for the edge subgraphSeeds
        for(auto it = distinctDataEdges.begin(); it != distinctDataEdges.end(); ++it){
            Subgraph subgraph;
            subgraph.edgeLabels.insert(std::make_pair(std::make_pair(0,1), it->first)); // add the pattern with nodes 0 and 1 |
subgraph.frequency = it->second;
            subgraph.multiEdges.insert(it->first);
            subgraphSeeds.push_back(subgraph);
        }
*/
        /// Collecting the sub-multiedges.

        std::set<std::set<int>> distinctEdges; // These edges are lexicographically ordered |
        CollectAllDataEdges(distinctDataEdges, distinctEdges);
//        CollectClosedDataEdges(distinctDataEdges, distinctEdges); // Use FP trie to retrieve only closed itemsets for edgeSet |
cout << "# distinct itemsets: " << distinctEdges.size() << endl;
cout << endl;
        /// Create structure for the edge subgraphSeeds
        for(auto it = distinctEdges.begin(); it != distinctEdges.end(); ++it){
            Subgraph subgraph;
            subgraph.edgeLabels.insert(std::make_pair(std::make_pair(0,1), *it)); // add the pattern with nodes 0 and 1 |
            subgraph.multiEdges.insert(*it);
            subgraphSeeds.push_back(subgraph);
        }

    }
}

void SubgraphMiner::ExtractInfo(const Subgraph& subgraph, GraphParameter& subgraphInfo)
{
/// Node labels exist
/*
    Update: subgraphInfo.attributes
    */
/// ###

    std::set<int> nodes;
    for(auto it = subgraph.edgeLabels.begin(); it != subgraph.edgeLabels.end(); ++it){
        nodes.insert(it->first.first);
        nodes.insert(it->first.second);
    }
    subgraphInfo.nNodes = nodes.size();

    subgraphInfo.eLabelMap = subgraph.edgeLabels; // add the original edges |
    for(auto it = subgraph.edgeLabels.begin(); it != subgraph.edgeLabels.end(); ++it){ // add the reversed edges, for the undirected graphs |
        std::pair<int, int> edgeRev = std::make_pair(it->first.second, it->first.first);
        subgraphInfo.eLabelMap.insert(std::make_pair(edgeRev, it->second));
    }

    Vector2D adjListTemp(subgraphInfo.nNodes);
    EdgeLabel nbrSignTemp(subgraphInfo.nNodes);
    for(auto it = subgraph.edgeLabels.begin(); it != subgraph.edgeLabels.end(); ++it){
        nbrSignTemp[it->first.first].push_back(it->second);
        nbrSignTemp[it->first.second].push_back(it->second);
        if (find(adjListTemp[it->first.first].begin(), adjListTemp[it->first.first].end(), it->first.second) == adjListTemp[it->first.first].end())
          adjListTemp[it->first.first].push_back(it->first.second);
        if (find(adjListTemp[it->first.second].begin(), adjListTemp[it->first.second].end(), it->first.first) == adjListTemp[it->first.second].end())
          adjListTemp[it->first.second].push_back(it->first.first);
    }
    subgraphInfo.adjacencyList = adjListTemp;
    subgraphInfo.neighbourSign = nbrSignTemp;
}

void SubgraphMiner::FindFrequentSeeds(const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newSubgraphs, std::deque<Subgraph>& newFSGs)
{
    for(size_t i = 0; i < newSubgraphs.size(); ++i){
        GraphParameter subgraphInfo;
        ExtractInfo(newSubgraphs[i], subgraphInfo);
        Sumgra subIsomorphism;

/// FInd support using MIn Image Support / Maximum Ind. Set separately
//        subIsomorphism.FindEmbeddings(supportVal, subgraphInfo, dataGraphInfo, graphIndexes, newSubgraphs[i]);
//        if(!newSubgraphs[i].embeddings.empty()){
////cout << "Initial Emb: " << newSubgraphs[i].embeddings.size() << endl;
//            Frequency supportMeasure;
//            if(supportMeasure.IsMinImageFrequent(newSubgraphs[i].embeddings, supportVal))
//                newFSGs.push_back(newSubgraphs[i]);
//        }

/// FIND  support using Maximum Ind. Set ON THE FLY by modifying sumgra.
        if(subIsomorphism.FindEmbeddingsMNI(supportVal, subgraphInfo, dataGraphInfo, graphIndexes, newSubgraphs[i]))
            newFSGs.push_back(newSubgraphs[i]);
//cout << "size: " << newSubgraphs[i].invalidCands.size()  << ": #" << newSubgraphs[i].invalidCands.begin()->second.size() << endl;
    }
}

bool SubgraphMiner::IsIsomorphicCurrent(Subgraph& newSubgraph, std::set<std::vector<int>>& canonicalSet, const IsoTestMap& dataGraph)
{
    // Call TestIsoInvariants to avoid generating canonical OR compute automorphic group |
    ExtractInfo(newSubgraph, newSubgraph.parameters);
    GraphIso subgraph;
//cout << "Can Gen begun" << endl;
    subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
//cout << "Can Gen ended" << endl;
    if(canonicalSet.find(newSubgraph.canonicalForm) == canonicalSet.end()){
        canonicalSet.insert(newSubgraph.canonicalForm);
        return false;
    }
    else
        return true;
}

bool SubgraphMiner::IsIsomorphicPrevious(const int& subgraphSize, const std::vector<int>& canonical, IsoMeasure& canonicalAllLevel)
{
//    if(subgraphSize == 2)
//        return true; // No need to check |
//    else{
        auto it = canonicalAllLevel.find(subgraphSize);
        if(it == canonicalAllLevel.end()){ // Create a new level and add canonical elements to it |
            std::set<std::vector<int>> tmp;
//cout << "New canonical level: " << subgraphSize << endl;
//for(size_t i = 0; i < canonical.size(); ++i)
//cout << canonical[i] << ", ";
//cout << endl;
            tmp.insert(canonical);
            canonicalAllLevel.insert(std::make_pair(subgraphSize, tmp));
            return false;
        }
        else{
//cout << "Old canonical level: " << subgraphSize << endl;
//for(size_t i = 0; i < canonical.size(); ++i)
//cout << canonical[i] << ", ";
//cout << endl;
//if(subgraphSize == 5)
//cout << "canonicalSet.size() " << it->second.size() << endl;
            if(it->second.find(canonical) == it->second.end()){ // The subgraph in the same level does not exist |
                it->second.insert(canonical);
                return false;
            }
            else
                return true; // Subgraph already found |
        }
//    }
}


inline bool NotAvalidExtension(const Set2D& nonFreqEdges, const std::set<int>& newEdge)
{
    for(auto it = nonFreqEdges.begin(); it != nonFreqEdges.end(); ++it)
        if(std::includes(newEdge.begin(), newEdge.end(), (*it).begin(), (*it).end()))
            return true;
    return false;
}

bool SubgraphMiner::IsFrequent(const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, Subgraph& newSubgraph)
{
    Sumgra subIsomorphism;

/// FInd support using MIn Image Support / Maximum Ind. Set separately.
//    subIsomorphism.FindEmbeddings(supportVal, newSubgraph.parameters, dataGraphInfo, graphIndexes, newSubgraph);
//    if(!newSubgraph.embeddings.empty()){
////cout << "Next Emb: " << newSubgraph.embeddings.size() << endl;
//        Frequency supportMeasure;
//        if(supportMeasure.IsMinImageFrequent(newSubgraph.embeddings, supportVal))
//            return true;
//    }
//    return false;

/// FIND  support using Maximum Ind. Set ON THE FLY by modifying sumgra.
    if(subIsomorphism.FindEmbeddingsMNI(supportVal, newSubgraph.parameters, dataGraphInfo, graphIndexes, newSubgraph)){
//cout <<  newSubgraph.parameters.nNodes << ": returned sumgra" << endl;
//if(newSubgraph.parameters.nNodes == 3){
//    for(auto it = newSubgraph.edgeLabels.begin(); it != newSubgraph.edgeLabels.end(); ++it)
//        cout << *it->second.begin() << ", "; cout << endl;
//}
//cout << "***" << endl;
        return true;
    }
    else{
//        cout << "returned sumgra FALSE" << endl;
        return false;
    }
}

bool ContainsEdge(Subgraph& A, Subgraph& B) /// Check if all the multiedges of A are contained in at least one multiedge of B
{
    bool contains = false;
    for(auto itA = A.multiEdges.begin(); itA != A.multiEdges.end(); ++itA){
        contains = false;
        for(auto itB = B.multiEdges.begin(); itB != B.multiEdges.end(); ++itB){
            if(std::includes((*itB).begin(), (*itB).end(), (*itA).begin(), (*itA).end())){
                contains = true;
                break;
            }
        }
        if(!contains)
            return false;
    }
    return contains;
}

bool IsPreviousNonFrequent(NFSG& nonFSG, Subgraph& newSubgraph, IndexType& newSubgraphInd)
{
/// Possible Optimizations:
// 1. Use Hopcroft-Karp algorithm instead of 'ContainsEdge'.
// 2. Use indexing techniques to store the non-frequent SG (NFSG) for each level level. Could use a tree structure where a NFSG is a node in the index-tree.

    Sumgra subIso;
//    IndexType graphIndexes;
//    subIso.BuildIndexes(newSubgraph.parameters, graphIndexes);

    for(auto it = nonFSG.begin(); it != nonFSG.end(); ++it){
/// OPT-1: Can not be a subgraph |
        if(it->first >= newSubgraph.edgeLabels.size())
            break;
//        continue;
/// #
        for(auto itS = it->second.begin(); itS != it->second.end(); ++itS){
/// OPT-2: A multiedge of NFSG is not contained in new subgraph.
            if(!ContainsEdge(*itS, newSubgraph))
                continue;
/// #
            Vector2D solns;
            if(subIso.FindIsoMatch(*itS, newSubgraphInd, solns))
                return true;
        }
    }

    return false;

/// Observations:
// OPT-1 speeds up this function by 30%
// OPT-2 speeds up this function by 25%
}

bool IsAlreadyFSG(std::deque<IndexType>& maximalFSG, Subgraph& newSubgraph)
{
    for(size_t i = 0; i < maximalFSG.size(); ++i){
        if(newSubgraph.parameters.nNodes < maximalFSG[i].neighborTrie.size()){
            Vector2D solns;
            Sumgra subIso;
            if(subIso.FindIsoMatch(newSubgraph, maximalFSG[i], solns))
                return true;
        }
    }
    return false;
}


void SubgraphMiner::ExtendWithNode(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval)
{
    for(auto it = oldFSG.nonFreqEdgesN.begin(); it != oldFSG.nonFreqEdgesN.end(); ++it){
        Set2D nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for node *it
        if(!nonFreqEdges.empty()) // Non-frequent edge extensions are available for this extension |
            if(NotAvalidExtension(nonFreqEdges, newEdge))
                continue; // try the next extension |

        std::pair<int, int> tmpEdgeID = std::make_pair(it->first, oldFSG.orderedNodes.size());
        EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
        if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
            edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge));
        Set2D multiEdges = oldFSG.multiEdges;
        multiEdges.insert(newEdge); // maintain the set of multiedges |

        Subgraph newSubgraph;
        newSubgraph.edgeLabels = edgeLabels;
        newSubgraph.multiEdges = multiEdges;
        ExtractInfo(newSubgraph, newSubgraph.parameters);
        GraphIso subgraph;
        subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
        if(IsAlreadyFSG(maximalFSG, newSubgraph))
            continue;
        if(IsIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//++timeEval.prevISOyes;
                continue;
        }
//        else
//++timeEval.prevISOno;

        Sumgra subIso;
        IndexType newSubgraphInd;
        subIso.BuildIndexes(newSubgraph.parameters, newSubgraphInd);

//clock_t tNFSG = clock();
            if(IsPreviousNonFrequent(nonFSG, newSubgraph, newSubgraphInd)){
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//++timeEval.nonFrequent;
                continue;
            }
//            else{
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//            }

        getInvalidCands(oldFSG, oldFSGs, fsgs, newSubgraph, newSubgraphInd);

//clock_t Time = clock();
        if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
            newFSGs.push_back(newSubgraph);
        }
        else{
            it->second.insert(newEdge);
            /// Collect non-frequent SGs
            auto itN = nonFSG.find(newSubgraph.edgeLabels.size());
            if(itN == nonFSG.end()){
                std::deque<Subgraph> tmp;
                tmp.push_back(newSubgraph);
                nonFSG.insert(std::make_pair(newSubgraph.edgeLabels.size(), tmp));
            }
            else{
                // Always ONLY a NEW non-frequent subgraph is added. No need to check again |
                itN->second.push_back(newSubgraph);
            }
        }
//timeEval.SgFrq += double(clock()-Time)/CLOCKS_PER_SEC;
//timeEval.EmbEv += newSubgraph.timeEval.EmbEv;
//timeEval.time1 += newSubgraph.timeEval.time1;
//timeEval.time2 += newSubgraph.timeEval.time2;
//timeEval.time3 += newSubgraph.timeEval.time3;
//timeEval.starPat += newSubgraph.timeEval.starPat;
//timeEval.normalPat += newSubgraph.timeEval.normalPat;
    }
}

void SubgraphMiner::ExtendWithoutNode(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval)
{
    for(auto it = oldFSG.nonFreqEdgesE.begin(); it != oldFSG.nonFreqEdgesE.end(); ++it){
        Set2D nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for edge *it
        if(!nonFreqEdges.empty()){ // Non-frequent edge extensions are available for this extension |
            if(NotAvalidExtension(nonFreqEdges, newEdge))
                continue; // try the next extension |
        }
        std::pair<int,int> newEdgeIDRev = std::make_pair(it->first.second, it->first.first);
        if(oldFSG.edgeLabels.find(it->first) != oldFSG.edgeLabels.end() || oldFSG.edgeLabels.find(newEdgeIDRev) != oldFSG.edgeLabels.end())
            continue;

        EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
        edgeLabels.insert(std::make_pair(it->first, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
        Set2D multiEdges = oldFSG.multiEdges;
        multiEdges.insert(newEdge); // maintain the set of multiedges |

        Subgraph newSubgraph;
        newSubgraph.edgeLabels = edgeLabels;
        newSubgraph.multiEdges = multiEdges;
        ExtractInfo(newSubgraph, newSubgraph.parameters);
        GraphIso subgraph;
        subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
        if(IsAlreadyFSG(maximalFSG, newSubgraph))
            continue;
        if(IsIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//++timeEval.prevISOyes;
            continue;
        }
//        else
//++timeEval.prevISOno;

        Sumgra subIso;
        IndexType newSubgraphInd;
        subIso.BuildIndexes(newSubgraph.parameters, newSubgraphInd);

//clock_t tNFSG = clock();
        if(IsPreviousNonFrequent(nonFSG, newSubgraph, newSubgraphInd)){
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//++timeEval.nonFrequent;
            continue;
        }
//        else{
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//        }

        getInvalidCands(oldFSG, oldFSGs, fsgs, newSubgraph, newSubgraphInd);

//clock_t Time = clock();

        if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
            newFSGs.push_back(newSubgraph);
        }
        else{
            it->second.insert(newEdge);
            /// Collect non-frequent SGs
            auto itN = nonFSG.find(newSubgraph.edgeLabels.size());
            if(itN == nonFSG.end()){
                std::deque<Subgraph> tmp;
                tmp.push_back(newSubgraph);
                nonFSG.insert(std::make_pair(newSubgraph.edgeLabels.size(), tmp));
            }
            else{
                // Always ONLY a NEW non-frequent subgraph is added. No need to check again |
                itN->second.push_back(newSubgraph);
            }
        }
//timeEval.SgFrq += double(clock()-Time)/CLOCKS_PER_SEC;
//timeEval.EmbEv += newSubgraph.timeEval.EmbEv;
//timeEval.time1 += newSubgraph.timeEval.time1;
//timeEval.time2 += newSubgraph.timeEval.time2;
//timeEval.time3 += newSubgraph.timeEval.time3;
//timeEval.starPat += newSubgraph.timeEval.starPat;
//timeEval.normalPat += newSubgraph.timeEval.normalPat;
    }
}

void SubgraphMiner::ExtendWithNodeNew(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval)
{
    Set2D temp;
    int nNodes = oldFSG.orderedNodes.size(); // no of nodes in the input subgraph |
    std::set<std::vector<int>> extraNodeCanonical;
    for(auto it = oldFSG.orderedNodes.begin(); it != oldFSG.orderedNodes.end(); ++it){
        EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
        std::pair<int, int> tmpEdgeID = std::make_pair(*it, nNodes);
        if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
            edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge)); // add a new edge: a new node "nNodes" forming an edge with an already existing node "*it" |
        Set2D multiEdges = oldFSG.multiEdges;
        multiEdges.insert(newEdge); // maintain the set of multiedges |

        Subgraph newSubgraph;
        newSubgraph.edgeLabels = edgeLabels;
        newSubgraph.multiEdges = multiEdges;
        if(IsIsomorphicCurrent(newSubgraph, extraNodeCanonical, dataGraph)){ // Check if this new set of edges "edgeLabels" are isomorphic to the already existing "newSubgraphs"
//++timeEval.currISOyes;
            continue;
        }
//        else
//++timeEval.currISOno;

        if(IsAlreadyFSG(maximalFSG, newSubgraph))
            continue;

        oldFSG.nonFreqEdgesN.insert(std::make_pair((*it), temp)); // This step assigns a set of unique nodes that are to be extended. Each node is selected on a first-come-first-serve basis, which is a representative node from its permutation group.

        if(IsIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//++timeEval.prevISOyes;
                continue;
        }
//        else
//++timeEval.prevISOno;

        Sumgra subIso;
        IndexType newSubgraphInd;
        subIso.BuildIndexes(newSubgraph.parameters, newSubgraphInd);

//clock_t tNFSG = clock();
            if(IsPreviousNonFrequent(nonFSG, newSubgraph, newSubgraphInd)){
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//++timeEval.nonFrequent;
                continue;
            }
//            else{
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//            }


        getInvalidCands(oldFSG, oldFSGs, fsgs, newSubgraph, newSubgraphInd);

//clock_t Time = clock();
        if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
            newFSGs.push_back(newSubgraph);
        }
        else{
            oldFSG.nonFreqEdgesN.find((*it))->second.insert(newEdge);
            /// Collect non-frequent SGs
            auto itN = nonFSG.find(newSubgraph.edgeLabels.size());
            if(itN == nonFSG.end()){
                std::deque<Subgraph> tmp;
                tmp.push_back(newSubgraph);
                nonFSG.insert(std::make_pair(newSubgraph.edgeLabels.size(), tmp));
            }
            else{
                // Always ONLY a NEW non-frequent subgraph is added. No need to check again |
                itN->second.push_back(newSubgraph);
            }
        }
//timeEval.SgFrq += double(clock()-Time)/CLOCKS_PER_SEC;
//timeEval.EmbEv += newSubgraph.timeEval.EmbEv;
//timeEval.time1 += newSubgraph.timeEval.time1;
//timeEval.time2 += newSubgraph.timeEval.time2;
//timeEval.time3 += newSubgraph.timeEval.time3;
//timeEval.starPat += newSubgraph.timeEval.starPat;
//timeEval.normalPat += newSubgraph.timeEval.normalPat;
    }
}


void SubgraphMiner::ExtendWithoutNodeNew(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval)
{
    Set2D temp;
    std::set<std::pair<int,int>> extendableEdges;
    std::set<std::vector<int>> noNodeCanonical;
    for(size_t i = 0; i < oldFSG.orderedNodes.size()-1; ++i){
        for(size_t j = i+1; j < oldFSG.orderedNodes.size(); ++j){
            std::pair<int,int> newEdgeID = std::make_pair(oldFSG.orderedNodes[i], oldFSG.orderedNodes[j]); // a possible edge in the subgraph clique |
            std::pair<int,int> newEdgeIDRev = std::make_pair(oldFSG.orderedNodes[j], oldFSG.orderedNodes[i]); // a possible edge in the subgraph clique |
            if(oldFSG.edgeLabels.find(newEdgeID) != oldFSG.edgeLabels.end() || oldFSG.edgeLabels.find(newEdgeIDRev) != oldFSG.edgeLabels.end())
                continue;
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            edgeLabels.insert(std::make_pair(newEdgeID, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |

            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
            if(IsIsomorphicCurrent(newSubgraph, noNodeCanonical, dataGraph)){ // Check if this new set of edges "edgeLabels" are isomorphic to the already existing "newSubgraphs"
//++timeEval.currISOyes;
                continue;
            }
//            else
//++timeEval.currISOno;

            if(IsAlreadyFSG(maximalFSG, newSubgraph))
                continue;

            oldFSG.nonFreqEdgesE.insert(std::make_pair(newEdgeID, temp)); // This step assigns a set of unique node-pairs to which an edge can be added. Each node-pair is selected on a first-come-first-serve basis, which is a representative node-pair from its permutation group.
            if(IsIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//++timeEval.prevISOyes;
                continue;
            }
//            else
//++timeEval.prevISOno;

        Sumgra subIso;
        IndexType newSubgraphInd;
        subIso.BuildIndexes(newSubgraph.parameters, newSubgraphInd);

//clock_t tNFSG = clock();
            if(IsPreviousNonFrequent(nonFSG, newSubgraph, newSubgraphInd)){
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//++timeEval.nonFrequent;
                continue;
            }
//            else{
//timeEval.SgExt += double(clock()-tNFSG)/CLOCKS_PER_SEC;
//            }

            getInvalidCands(oldFSG, oldFSGs, fsgs, newSubgraph, newSubgraphInd);

//clock_t Time = clock();

            if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                newFSGs.push_back(newSubgraph);
            }
            else{
                oldFSG.nonFreqEdgesE.find(newEdgeID)->second.insert(newEdge);
                /// Collect non-frequent SGs
                auto itN = nonFSG.find(newSubgraph.edgeLabels.size());
                if(itN == nonFSG.end()){
                    std::deque<Subgraph> tmp;
                    tmp.push_back(newSubgraph);
                    nonFSG.insert(std::make_pair(newSubgraph.edgeLabels.size(), tmp));
                }
                else{
                    // Always ONLY a NEW non-frequent subgraph is added. No need to check again |
                    itN->second.push_back(newSubgraph);
                }
            }
//timeEval.SgFrq += double(clock()-Time)/CLOCKS_PER_SEC;
//timeEval.EmbEv += newSubgraph.timeEval.EmbEv;
//timeEval.time1 += newSubgraph.timeEval.time1;
//timeEval.time2 += newSubgraph.timeEval.time2;
//timeEval.time3 += newSubgraph.timeEval.time3;
//timeEval.starPat += newSubgraph.timeEval.starPat;
//timeEval.normalPat += newSubgraph.timeEval.normalPat;
        }
    }
}

/// USING SET DATA STRUCTURE
void SubgraphMiner::SubgraphExtension(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::deque<Subgraph>& newFSGs, std::deque<Subgraph> oldFSGs, std::deque<IndexType>& maximalFSG, FSGs& fsgs, TimeEval& timeEval)
{
    if(oldFSG.alreadyExtended){ // If the subgraph has already been extended |
        // Ext 1. Adding an edge WITH an extra node.
        ExtendWithNode(dataGraph, oldFSG, newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, newFSGs, oldFSGs, maximalFSG, fsgs, timeEval);
        // Ext 2. Adding an edge WITHOUT an extra node.
        ExtendWithoutNode(dataGraph, oldFSG, newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, newFSGs, oldFSGs, maximalFSG, fsgs, timeEval);
    }
    else{
        // Ext 1. Adding an edge WITH an extra node.
        ExtendWithNodeNew(dataGraph, oldFSG, newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, newFSGs, oldFSGs, maximalFSG, fsgs, timeEval);
        // Ext 2. Adding an edge WITHOUT an extra node.
        ExtendWithoutNodeNew(dataGraph, oldFSG, newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, newFSGs, oldFSGs, maximalFSG, fsgs, timeEval);
        // Comment the below line to UNPRUNE the unfrequent-edge-type-extension approach.
        oldFSG.alreadyExtended = true; // "oldFSG" has just been extended.
    }
}


/// USING TRIE DATA STrUCTURE

/*

void SubgraphMiner::SubgraphExtension(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newFSGs)
{
/// Return only the frequent subgraphs.
//std::vector<int> compare = {0,8,8};
//Ext 1: No of edges to be added = no. of nodes present in the subgraph. However, they contain isomorphic subgraphs |
//Ext 2: No of nodes to be added = Clique edges - existing edges in the subgraph. However, they may contain isomorphic subgraphs |

//    std::set<int> newEdge = fsgExtened.edgeLabels.begin()->second; // a new edge to be added |
//cout << "Extension started" << endl;
    if(oldFSG.alreadyExtended){ // If the subgraph has already been extended |
//cout << "already extended pattern" << endl;
        /// Ext 1. Adding an edge WITH an extra node.
        for(auto it = oldFSG.nonFreqEdgeN.begin(); it != oldFSG.nonFreqEdgeN.end(); ++it){
            Trie* nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for node *it
            if(nonFreqEdges->Root()->children().size() != 0){ // Non-frequent edge extensions are available for this extension |
                if(InvalidTrieEdge(nonFreqEdges, newEdge)){
//if(oldFSG.canonicalForm == compare){
//cout << "Preorder BEFORE: " << it->first << ": " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "New edge: ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//}
                    continue; // try the next extension |
                }
            }
            std::pair<int, int> tmpEdgeID = std::make_pair(it->first, oldFSG.orderedNodes.size());
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
                edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge));
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |
            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
            ExtractInfo(newSubgraph, newSubgraph.parameters);
            GraphIso subgraph;
            subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
            if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
                if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                    newFSGs.push_back(newSubgraph);
//        cout << "returned sumgra 1" << endl;

                }
                else{
//        cout << "returned sumgra 1" << endl;

//if(oldFSG.canonicalForm == compare){
//cout << "Preorder BEFORE: " << it->first << ": " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//}
                    it->second->AddNonFreqEdges(newEdge);
//if(oldFSG.canonicalForm == compare){
//cout << "Preorder AFTER: " << it->first << ": " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//}
                }
            }
        }
        /// Ext 2. Adding an edge WITHOUT an extra node.
//if(compare == oldFSG.canonicalForm && newEdge.size() == 2){
//cout << "HERE : " << oldFSG.nonFreqEdgeE.size()  << endl;
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//Trie* nonFreqEdges = oldFSG.nonFreqEdgeE.begin()->second;
//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//}

        for(auto it = oldFSG.nonFreqEdgeE.begin(); it != oldFSG.nonFreqEdgeE.end(); ++it){
            Trie* nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for edge *it
//            if(nonFreqEdges->Root()->contentInt() == 0){ // Non-frequent edge extensions are available for this extension |
            if(nonFreqEdges->Root()->children().size() != 0){
//                if(NotAvalidExtension(nonFreqEdges, newEdge))
                if(InvalidTrieEdge(nonFreqEdges, newEdge)){
//if(compare == oldFSG.canonicalForm){
//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//cout << "New edge: ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn+1 << ",";
//cout << endl << "***" << endl;
//}

//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//                    continue; // try the next extension |
                }
            }
            std::pair<int,int> newEdgeIDRev = std::make_pair(it->first.second, it->first.first);

            if(oldFSG.edgeLabels.find(it->first) == oldFSG.edgeLabels.end() && oldFSG.edgeLabels.find(newEdgeIDRev) == oldFSG.edgeLabels.end()){
                EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
                edgeLabels.insert(std::make_pair(it->first, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
                Set2D multiEdges = oldFSG.multiEdges;
                multiEdges.insert(newEdge); // maintain the set of multiedges |
                Subgraph newSubgraph;
                newSubgraph.edgeLabels = edgeLabels;
                newSubgraph.multiEdges = multiEdges;
                ExtractInfo(newSubgraph, newSubgraph.parameters);
                GraphIso subgraph;
                subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
                if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
                    if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
//        cout << "returned sumgra 2" << endl;

                        newFSGs.push_back(newSubgraph);
//if(compare == oldFSG.canonicalForm)
//cout << "frequent " << newFSGs.size() << endl;

                    }
                    else{
//        cout << "returned sumgra 2" << endl;

//                        cout << "???" << endl;
                        it->second->AddNonFreqEdges(newEdge);
//if(compare == oldFSG.canonicalForm){
//Trie* nonFreqEdges = it->second;
//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//cout << "New edge: ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn+1 << ",";
//cout << endl << "***" << endl;
//}
                    }
                }
            }
        }
    }
    else{ // The subgraph "oldFSG" is being extended the first time |
//cout << "Initial extension" << endl;
//        newTrie->Root()->setContentInt(-1);
//if(oldFSG.edgeLabels.size() == 1 && oldFSG.edgeLabels.begin()->second.size() == 2)
//    cout << "HELLLOOOOOOO: " << *oldFSG.edgeLabels.begin()->second.begin() << ": " << *newEdge.begin() << endl;
        /// Ext 1. Adding an edge WITH an extra node.
        int nNodes = oldFSG.orderedNodes.size(); // no of nodes in the input subgraph |
        std::set<std::vector<int>> extraNodeCanonical;
        for(auto it = oldFSG.orderedNodes.begin(); it != oldFSG.orderedNodes.end(); ++it){
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            std::pair<int, int> tmpEdgeID = std::make_pair(*it, nNodes);
            if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
                edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge)); // add a new edge: a new node "nNodes" forming an edge with an already existing node "*it" |
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |
            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
//cout << "Check current ISO 3 " << edgeLabels.size() << endl;
            if(IsNotIsomorphicCurrent(newSubgraph, extraNodeCanonical, dataGraph)){ // Check if this new set of edges "edgeLabels" are isomorphic to the already existing "newSubgraphs"
//cout << " current ISO checked 3" << endl;

                Trie* newTrie = new Trie();
                oldFSG.nonFreqEdgeN[(*it)] = newTrie;
//cout << "Check previous ISO 3" << endl;
                if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//cout << "ISO checked 3" << endl;
                    if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
//        cout << "returned sumgra 3" << endl;

                        newFSGs.push_back(newSubgraph);
                    }
                    else{
//        cout << "returned sumgra 3" << endl;

                        oldFSG.nonFreqEdgeN.find((*it))->second->AddNonFreqEdges(newEdge);
                    }
                }
            }
        }
        /// Ext 2. Adding an edge WITHOUT an extra node.
        std::set<std::pair<int,int>> extendableEdges;
        std::set<std::vector<int>> noNodeCanonical;
        for(size_t i = 0; i < oldFSG.orderedNodes.size()-1; ++i){
            for(size_t j = i+1; j < oldFSG.orderedNodes.size(); ++j){
                std::pair<int,int> newEdgeID = std::make_pair(oldFSG.orderedNodes[i], oldFSG.orderedNodes[j]); // a possible edge in the subgraph clique |
                std::pair<int,int> newEdgeIDRev = std::make_pair(oldFSG.orderedNodes[j], oldFSG.orderedNodes[i]); // a possible edge in the subgraph clique |
                if(oldFSG.edgeLabels.find(newEdgeID) == oldFSG.edgeLabels.end() && oldFSG.edgeLabels.find(newEdgeIDRev) == oldFSG.edgeLabels.end()){
                    EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
                    edgeLabels.insert(std::make_pair(newEdgeID, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
                    Set2D multiEdges = oldFSG.multiEdges;
                    multiEdges.insert(newEdge); // maintain the set of multiedges |
                    Subgraph newSubgraph;
                    newSubgraph.edgeLabels = edgeLabels;
                    newSubgraph.multiEdges = multiEdges;
//cout << "Check current ISO 4 " << edgeLabels.size() << endl;
                    if(IsNotIsomorphicCurrent(newSubgraph, noNodeCanonical, dataGraph)){ // Check if this new set of edges "edgeLabels" are isomorphic to the already existing "newSubgraphs"
//cout << " current ISO checked 4" << endl;

                        Trie* newTrie = new Trie();
                        oldFSG.nonFreqEdgeE[newEdgeID] = newTrie;//if(compare == oldFSG.canonicalForm)//if(compare == oldFSG.canonicalForm)
//cout << "Check previous ISO 4" << endl;
                        if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//cout << "ISO checked 4" << endl;

//                            oldFSG.nonFreqEdgeE[newEdgeID] = newTrie;
                            if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
//        cout << "returned sumgra 4 " << endl;

                                newFSGs.push_back(newSubgraph);
                            }
                            else{
//        cout << "returned sumgra 4"  << endl;

                                oldFSG.nonFreqEdgeE.find(newEdgeID)->second->AddNonFreqEdges(newEdge);
                            }
                        }
                    }
                }
            }
        }
        oldFSG.alreadyExtended = true; // "oldFSG" has just been extended.
    }
//    cout << "Exit EXtension" << endl;
}
*/

/*
void SubgraphMiner::FindFrequentSubgraphs(TimeEval& timeEval, const std::set<int> newEdge, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, Subgraph& oldFSG, std::deque<Subgraph>& newSubgraphs, std::deque<Subgraph>& newFSGs)
{
    for(size_t i = 0; i < newSubgraphs.size(); ++i){
        Sumgra subIsomorphism;
            clock_t Time = clock();
//        cout << "hello 1" <<  endl;
//        cout << newSubgraphs[i].parameters.eLabelMap.size() << endl;
        subIsomorphism.FindEmbeddings(supportVal, newSubgraphs[i].parameters, dataGraphInfo, graphIndexes, newSubgraphs[i]);
//        cout << "hello 2" <<  endl;

            timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
        if(!newSubgraphs[i].embeddings.empty()){
            Frequency supportMeasure;
            if(supportMeasure.IsMinImageFrequent(newSubgraphs[i].embeddings, supportVal)){
                newFSGs.push_back(newSubgraphs[i]);
            }
            else{
                oldFSG.nonFreqEdgeExists = true;
                oldFSG.nonFreqEdges.insert(newEdge);
                oldFSG.nonFreqEdge = new Trie();
                oldFSG.nonFreqEdge->AddMultiedges(newEdge);
//for(auto it = newEdge.begin(); it != newEdge.end(); ++it)
//    cout << *it << ",";
//cout << endl;
            }
        }
        else{
            oldFSG.nonFreqEdgeExists = true;
            oldFSG.nonFreqEdges.insert(newEdge);
            oldFSG.nonFreqEdge = new Trie();
            oldFSG.nonFreqEdge->AddMultiedges(newEdge);
//for(auto it = newEdge.begin(); it != newEdge.end(); ++it)
//    cout << *it << ",";
//cout << endl;
        }
    }
}
*/

void CollectFSG(const int& sSize, const std::deque<Subgraph> newFSGs, std::map<int, std::deque<FSG>>& allFSG)
{
    std::deque<FSG> newFSG;
    for(size_t i = 0; i < newFSGs.size(); ++i){
        FSG iNewFSG;
        iNewFSG.structure = newFSGs[i].edgeLabels;
        newFSG.push_back(iNewFSG);
    }
    auto it = allFSG.find(sSize);
    if(it == allFSG.end())
        allFSG.insert(std::make_pair(sSize, newFSG));
    else
        it->second.insert(it->second.end(), newFSG.begin(), newFSG.end());
}

inline bool IsvalidEdge(const Subgraph& nextSubgraph, const std::set<int>& newEdge)
{
    for(auto it = nextSubgraph.nonFreqEdges.begin(); it != nextSubgraph.nonFreqEdges.end(); ++it)
        if(std::includes(newEdge.begin(), newEdge.end(), (*it).begin(), (*it).end()))
            return false;
    return true;
}


inline bool SubgraphMiner::InvalidTrieEdge(Trie* trie, std::set<int> newEdge)
{
    std::set<int> newEdgeTemp = newEdge;
    /// The for loop is necessary, so as to accomplish the superset containment.
/// Work on "IfEdgeExists" function; it does not execute accurately yet.
    for(auto it = newEdge.begin(); it != newEdge.end(); ++it){
        bool found = false;
        int pos = 0;
//        if(trie->EdgeExists(trie->Root(), newEdgeTemp, pos, found))
//            return true;
        trie->EdgeExists(trie->Root(), newEdgeTemp, pos, found);
        if(found)
            return true;

        newEdgeTemp.erase(newEdgeTemp.begin());
    }
    return false;
}

void SubgraphMiner::CoPExtension(StackFSG& fsgClassStack, IsoMeasure& canonicalAllLevel, std::deque<Subgraph>& extendableEdges, int nE, TimeEval& timeEval, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg, std::deque<IndexType>& maximalFSG, FSGs& fsgs, int& tg)
{
    if (nE == extendableEdges.size())
        return; // go to next extension |
    while(!fsgClassStack.empty()){ // iterate until the entire DFS tree is spanned |

//        std::deque<Subgraph> newFSGs;
//        std::set<int> newEdge = extendableEdges[nE].edgeLabels.begin()->second;
//        SubgraphExtension(dataGraph, fsgClassStack.front().front(), newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, newFSGs, fsgClassStack.front(), maximalFSG, fsgs, timeEval);

        std::deque<Subgraph> newFSGs;
        std::set<int> newEdge = extendableEdges[nE].edgeLabels.begin()->second;
//cout << "Size: " << fsgClassStack.front().size() << endl;
        for(size_t i = 0; i < fsgClassStack.front().size(); ++i){
            std::deque<Subgraph> partialNewFSGs;
            SubgraphExtension(dataGraph, fsgClassStack.front()[i], newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, partialNewFSGs, fsgClassStack.front(), maximalFSG, fsgs, timeEval);
            newFSGs.insert(newFSGs.end(), partialNewFSGs.begin(), partialNewFSGs.end());
        }
//cout << "NEW: " << newFSGs.size() << endl;

        if (!newFSGs.empty()){ // Grow the DFS tree until we find frequent subgraphs ; else go back to the next stack element |
            fsgClassStack.push_front(newFSGs); // Add the new FSG to the stack |
            auto itN = fsgs.find(newFSGs.front().edgeLabels.size());
            if(itN == fsgs.end()){
                fsgs.insert(std::make_pair(newFSGs.front().edgeLabels.size(), newFSGs));
            }
            else{
                itN->second.insert(itN->second.end(), newFSGs.begin(), newFSGs.end());
            }
            CollectFSG(newFSGs.front().edgeLabels.size(), newFSGs, allFSG);
            noOfFsg += newFSGs.size(); // count all the subgraphs |
        }
        else{ // Adding "newEdge" to "oldFSG" did not produce newFSGs |
            CoPExtension(fsgClassStack, canonicalAllLevel, extendableEdges, nE+1, timeEval, dataGraph, supportVal, dataGraphInfo, graphIndexes, nonFSG, allFSG, noOfFsg, maximalFSG, fsgs, tg);

            if(!fsgClassStack.empty()){
                /// Collect maximal patterns [Efficient without it].

                if(fsgClassStack.size() >= tg){ // check if it is maximal pattern |
                    tg = fsgClassStack.size();
                    if(fsgClassStack.front().front().edgeLabels.size() > 2){ // check if the pattern has at least 3 edges |
                        // store only the index structure of maximal patterns |

                        Sumgra subIso;
                        IndexType FSGindex;
                        subIso.BuildIndexes(fsgClassStack.front().front().parameters, FSGindex);
                        maximalFSG.push_back(FSGindex);

                    }
                }
                fsgClassStack.pop_front();
//                if(fsgClassStack.front().size() == 1){
//                    fsgClassStack.pop_front(); // remove the entire "deque"
//                }
//                else{
//                    fsgClassStack.front().pop_front(); // remove only one deque-element
//                }
            }
        }
    }
}


void SubgraphMiner::DfsRecursiveMining(StackFSG& fsgClassStack, IsoMeasure& canonicalAllLevel, std::deque<Subgraph>& extendableEdges, int nE, TimeEval& timeEval, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& nonFSG, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg, std::deque<IndexType>& maximalFSG, FSGs& fsgs, int& tg)
{
    if (nE == extendableEdges.size())
        return; // go to next extension |

    while(!fsgClassStack.empty()){ // iterate until the entire DFS tree is spanned |
        std::deque<Subgraph> newFSGs;
        std::set<int> newEdge = extendableEdges[nE].edgeLabels.begin()->second;

        SubgraphExtension(dataGraph, fsgClassStack.front().front(), newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, nonFSG, newFSGs, fsgClassStack.front(), maximalFSG, fsgs, timeEval);

        if (!newFSGs.empty()){ // Grow the DFS tree until we find frequent subgraphs ; else go back to the next stack element |
            fsgClassStack.push_front(newFSGs); // Add the new FSG to the stack |
            auto itN = fsgs.find(newFSGs.front().edgeLabels.size());
            if(itN == fsgs.end()){
                fsgs.insert(std::make_pair(newFSGs.front().edgeLabels.size(), newFSGs));
            }
            else{
                itN->second.insert(itN->second.end(), newFSGs.begin(), newFSGs.end());
            }
            CollectFSG(newFSGs.front().edgeLabels.size(), newFSGs, allFSG);
            noOfFsg += newFSGs.size(); // count all the subgraphs |
        }
        else{ // Adding "newEdge" to "oldFSG" did not produce newFSGs |
            DfsRecursiveMining(fsgClassStack, canonicalAllLevel, extendableEdges, nE+1, timeEval, dataGraph, supportVal, dataGraphInfo, graphIndexes, nonFSG, allFSG, noOfFsg, maximalFSG, fsgs, tg);

            if(!fsgClassStack.empty()){
                /// Collect maximal patterns [Efficient without it].

                if(fsgClassStack.size() >= tg){ // check if it is maximal pattern |
                    tg = fsgClassStack.size();
                    if(fsgClassStack.front().front().edgeLabels.size() > 2){ // check if the pattern has at least 3 edges |
                        // store only the index structure of maximal patterns |

                        Sumgra subIso;
                        IndexType FSGindex;
                        subIso.BuildIndexes(fsgClassStack.front().front().parameters, FSGindex);
                        maximalFSG.push_back(FSGindex);

                    }
                }

                if(fsgClassStack.front().size() == 1){
                    fsgClassStack.pop_front(); // remove the entire "deque"
                }
                else{
                    fsgClassStack.front().pop_front(); // remove only one deque-element
                }
            }

        }
        nE = 0; // reset the pointer for next Extension - nE |
    }
}


void SubgraphMiner::ArrangeByCoP(std::deque<Subgraph>& fsgSeedsSorted, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, NFSG& basicNonFSG, std::deque<IndexType>& maximalFSG, FSGs& fsgs, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg, CoP& cop, TimeEval& timeEval)
{
    for (size_t iS = 0; iS < fsgSeedsSorted.size(); ++iS) {
        std::deque<Subgraph> extendableEdges(fsgSeedsSorted.begin()+iS, fsgSeedsSorted.end());
        IsoMeasure canonicalAllLevel; // For the DFS subtree that grows from "fsgSeedsSorted[iS]" |
        for (size_t nE = 0; nE < extendableEdges.size(); ++nE) {
            std::deque<Subgraph> newFSGs;
            std::set<int> newEdge = extendableEdges[nE].edgeLabels.begin()->second;
            std::deque<Subgraph> siblings; // only one in this case |
            siblings.push_back(fsgSeedsSorted[iS]);
            SubgraphExtension(dataGraph, fsgSeedsSorted[iS], newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, basicNonFSG, newFSGs, siblings, maximalFSG, fsgs, timeEval); // level 2
            if (!newFSGs.empty()){
                CollectFSG(2, newFSGs, allFSG);
                noOfFsg += newFSGs.size(); // count all the subgraphs |
                /// extend further.
                for (size_t nEE = 0; nEE < extendableEdges.size(); ++nEE) {
                    std::deque<Subgraph> newFSGs3;
                    std::set<int> newEdge = extendableEdges[nEE].edgeLabels.begin()->second;
                    SubgraphExtension(dataGraph, newFSGs.front(), newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, basicNonFSG, newFSGs3, newFSGs, maximalFSG, fsgs, timeEval); // level 3
                    if (!newFSGs3.empty()){
                        CollectFSG(3, newFSGs3, allFSG);
                        noOfFsg += newFSGs3.size(); // count all the subgraphs |
                        std::multiset<std::set<int>> eLabels;
                        for(auto it = newFSGs3.front().edgeLabels.begin(); it != newFSGs3.front().edgeLabels.end(); ++it)
                            eLabels.insert(it->second);
                        auto itM = cop.find(eLabels);
                        if(itM == cop.end()){
                            cop.insert(std::make_pair(eLabels, newFSGs3));
                        }
                        else{
                            itM->second.insert(itM->second.begin(), newFSGs3.begin(), newFSGs3.end());
                        }
                    }
                }
            }
        }
    }
}


void SubgraphMiner::MineFrequentSubgraphs(const GraphParameter& dataGraphInfo,  IndexType& graphIndexes, const int& supportVal, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg)
{
    /// Create initial seeds for mining.
    std::deque<Subgraph> subgraphSeeds, fsgSeeds;
    CollectSubgraphSeeds(dataGraphInfo, subgraphSeeds);
    cout << "Subgraph seeds: " << subgraphSeeds.size() << endl;
    FindFrequentSeeds(supportVal, dataGraphInfo, graphIndexes, subgraphSeeds, fsgSeeds);
    cout << "Frequent seeds: " << fsgSeeds.size() << endl;


    /// Collect the discovered FSG seed for outputting.
    CollectFSG(1, fsgSeeds, allFSG);
    std::deque<FSG> frequentSubgraphs; // for better ordering of patterns use above data structure |
    noOfFsg = fsgSeeds.size(); // initialize the frequent subgraphs to the number of size-1 frequent subgraphs |

    /// Maintain a multiedge and vertex attribute mapping required for canonical representation of frequent subgraphs.
    IsoTestMap dataGraph;
    std::vector<int> edgeSize(fsgSeeds.size());
    for(size_t i = 0; i < fsgSeeds.size(); ++i){
        std::set<int> freqEdge = fsgSeeds[i].edgeLabels.begin()->second;
        dataGraph.multiedgeMap.insert(std::make_pair(freqEdge, i+1));
        edgeSize[i] = freqEdge.size();
    }

    /// Order the frequent seeds with their increasing size.

    std::vector<int> ind = SortIndexIncr(edgeSize);
    std::deque<Subgraph> fsgSeedsSorted(fsgSeeds.size());
    for(size_t i = 0; i < fsgSeeds.size(); ++i)
        fsgSeedsSorted[i] = fsgSeeds[ind[i]];


    /// Order the frequent seeds with their frequencies
/*
    std::vector<int> freqValue(fsgSeeds.size());
    for(size_t i = 0; i < fsgSeeds.size(); ++i)
        freqValue[i] = fsgSeeds[i].frequency;
    std::vector<int> ind = SortIndexIncr(freqValue);
    std::deque<Subgraph> fsgSeedsSorted(fsgSeeds.size());
    for(size_t i = 0; i < fsgSeeds.size(); ++i)
        fsgSeedsSorted[i] = fsgSeeds[ind[i]];
*/

/// Display
/*
for(size_t i = 0; i < fsgSeedsSorted.size(); ++i){
    std::set<int> freqEdge = fsgSeedsSorted[i].edgeLabels.begin()->second;
    for(auto it = freqEdge.begin(); it != freqEdge.end(); ++it)
        cout << *it << ",";

//    cout << ": " << fsgSeedsSorted[i].frequency << endl;
}
cout << endl;
*/
    std::deque<IndexType> maximalFSG;
    TimeEval timeEval;
    /// Used to find invalid candidates.
    FSGs fsgs;


/// Search space exploration WITH class of patterns (CoP).

    std::map<std::multiset<std::set<int>>, std::deque<Subgraph>> cop; // basic class of patterns of size 3 |
    NFSG basicNonFSG;
    ArrangeByCoP(fsgSeedsSorted, dataGraph, supportVal, dataGraphInfo, graphIndexes, basicNonFSG, maximalFSG, fsgs, allFSG, noOfFsg, cop, timeEval);

    int itr = 1;
    for(auto it = cop.begin(); it != cop.end(); ++it){
cout << "Iteration: " << itr << endl;
        std::vector<int> edges;
        for (size_t iS = 0; iS < fsgSeedsSorted.size(); ++iS) {
            if(it->first.find(fsgSeedsSorted[iS].edgeLabels.begin()->second) != it->first.end())
                edges.push_back(iS);
        }
        std::vector<int> eS = SortIndex(edges);
        std::deque<Subgraph> extendableEdges(fsgSeedsSorted.begin()+eS[0], fsgSeedsSorted.end());
        IsoMeasure canonicalAllLevel; // For the DFS subtree that grows from "fsgSeedsSorted[iS]" |
        StackFSG fsgClassStack; // FSG-class stack for DFS tree |
        fsgClassStack.push_front(it->second);
        NFSG nonFSG = basicNonFSG;
        int tg = 0, nE = 0;
        CoPExtension(fsgClassStack, canonicalAllLevel, extendableEdges, nE, timeEval, dataGraph, supportVal, dataGraphInfo, graphIndexes, nonFSG, allFSG, noOfFsg, maximalFSG, fsgs, tg);
        ++itr;
    }


/// Search space exploration WITHOUT class of patterns (CoP).
/*
    for (size_t iS = 0; iS < fsgSeedsSorted.size(); ++iS) { // iS -> subgraph iterator |
cout << "Iteration: " << iS << endl;
        /// Add only the lexicographically GE (>=) elements.
        std::deque<Subgraph> extendableEdges(fsgSeedsSorted.begin()+iS, fsgSeedsSorted.end());
        /// Add all elements.
//        std::deque<Subgraph> extendableEdges = fsgSeedsSorted;
        IsoMeasure canonicalAllLevel; // For the DFS subtree that grows from "fsgSeedsSorted[iS]" |
        StackFSG fsgClassStack; // FSG-class stack for DFS tree |
        fsgClassStack.push_front(std::deque<Subgraph>(fsgSeedsSorted.begin()+iS, fsgSeedsSorted.begin()+iS+1));
        NFSG nonFSG;
        int tg = 0;
        int nE = 0;
        DfsRecursiveMining(fsgClassStack, canonicalAllLevel, extendableEdges, nE, timeEval, dataGraph, supportVal, dataGraphInfo, graphIndexes, nonFSG, allFSG, noOfFsg, maximalFSG, fsgs, tg);
    }

*/

cout << endl <<"Maximal FSGs: " << maximalFSG.size() << endl;
//cout << "FSG levels = " << fsgs.size() << endl;
    /// Output all FSGs to file.
/*
    std::ofstream oFile;
    oFile.open("/home/vijay/Phd/MULTIGRAPHS/trunk/dataset/mining/testM/res1.txt");
    for(auto it = allFSG.begin(); it != allFSG.end(); ++it){
        oFile << "Level " << it->first << ": has " << it->second.size() << " patterns." << endl;
        for(size_t i  = 0; i < it->second.size(); ++i){
//            oFile << "Pattern " << i << endl;
            for(auto itm = it->second[i].structure.begin(); itm != it->second[i].structure.end(); ++itm){
                oFile << itm->first.first << ", " << itm->first.second << ": ";
                for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
                    oFile << *itn << "";
                    oFile << " E;  ";
            }
            oFile << endl;
        }
        oFile << endl;
    }
    oFile.close();
*/

    /// Output all FSGs to screen.
/*
    for(auto it = allFSG.begin(); it != allFSG.end(); ++it){
        cout << "Level: " << it->first << " has " << it->second.size() << " patterns." << endl;
        for(size_t i  = 0; i < it->second.size(); ++i){
            cout << "Pattern " << i << endl;
            for(auto itm = it->second[i].structure.begin(); itm != it->second[i].structure.end(); ++itm){
                for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
                    cout << *itn << ", ";
                    cout << " Edge; ";
            }
            cout << endl;
        }
        cout << endl;
    }
*/


    cout << endl << "No Of All FSG: " << noOfFsg << endl << endl;

/*
    cout << "Internal SUMGRA time:      " << timeEval.EmbEv << endl;
    cout << "Checking NFSG time:        " << timeEval.SgExt << endl;
    cout << "% NFSG computation time:   " << timeEval.SgExt/timeEval.SgFrq*100 << " %" << endl << endl;


    cout << "IsFreq:  " << timeEval.SgFrq << endl;
    cout << "time 1:  " << timeEval.time1 << endl;
    cout << "time 2:  " << timeEval.time2 << endl;
    cout << "time 3:  " << timeEval.time3 << endl << endl;


    cout << "ISO Current YES:  (No change with NFSG)   " << timeEval.currISOyes << endl;
    cout << "ISO Current NO:   (No change with NFSG)   " << timeEval.currISOno << endl;
    cout << "ISO Previous YES: (No change with NFSG)   " << timeEval.prevISOyes << endl;
    cout << "ISO Previous NO :  (Changes with NFSG)    " << timeEval.prevISOno << endl;
    cout << "ISO Previous NO REDUCTION:                " << timeEval.nonFrequent << endl;
    cout << "% of support check (sumgra) avoided:      " << 100*timeEval.nonFrequent/(timeEval.nonFrequent + timeEval.prevISOno) << " %" << endl << endl;


    cout << endl << "Star patterns: " << timeEval.starPat << endl;
    cout << "Norm patterns: " << timeEval.normalPat << endl;
*/
}

/*
1. Use invariants for iso test on current and previous [TestInvariants] (# nodes, #edges, #edge types, edge type equality)
2. Implement sub Iso, that will be used for
 2.1 passing pruning from all the siblings of one level to their child
 2.2 checking a new pattern against a MAXIMAL FREQUENT pattern
 2.3 checking a new pattern against a set of infrequent patterns already discovered.

eMandi
*/
