#include "SubgraphMinerNauty.h"

SubgraphMinerNauty::SubgraphMinerNauty()
{
    //ctor
}

SubgraphMinerNauty::~SubgraphMinerNauty()
{
    //dtor
}

/// Decompose the set into all possible (2^n-1) subsets and update to the collection of all multiedges.
void SubgraphMinerNauty::RetrieveAllItems(std::set<int> inputSet, VecOfSet& allSubsets)
{
    /// Initilize the solution |
    std::set<int> init;
    auto it = inputSet.begin();
    init.insert(*it);
    allSubsets.push_back(init);

    if(inputSet.size() > 1){
        std::advance(it, 1);
        for(; it != inputSet.end(); ++it){
            /// Add the new current element as a new subset|
            std::set<int> single;
            single.insert(*it);
            allSubsets.push_back(single);
            /// Grow the already existing subsets with the new element |
            VecOfSet alreadyCollected = allSubsets;
            for(size_t k = 0; k < alreadyCollected.size(); ++k){
                alreadyCollected[k].insert(*it);
                allSubsets.push_back(alreadyCollected[k]);
            }
        }
    }
}

void SubgraphMinerNauty::CollectAllDataEdges(const std::map<std::set<int>, int>& distinctDataEdges, std::set<std::set<int>>& distinctEdges)
{
    for(auto it = distinctDataEdges.begin(); it != distinctDataEdges.end(); ++it){
        VecOfSet allSubsets;
        RetrieveAllItems(it->first, allSubsets);
        for(auto itS = allSubsets.begin(); itS != allSubsets.end(); ++itS)
            distinctEdges.insert(*itS);
    }
}



void SubgraphMinerNauty::CollectSubgraphSeeds(const GraphParameter& dataGraphInfo, std::deque<Subgraph>& subgraphSeeds)
{
    if(dataGraphInfo.nodeLabelsExist){ // vertex labels are accounted for |
        cout << "code unwritten!" << endl;
    }
    else{ // no vertex labels |
        /// Collect all distinct multiedges in the datagraph and count their number of occurrences.
//        Trie * edgeTrie = new Trie(); // Maintain a multiedge trie structure |
        std::map<std::set<int>, int> distinctDataEdges; // <multiedge, its frequency>
        for(auto it = dataGraphInfo.eLabelMap.begin(); it != dataGraphInfo.eLabelMap.end(); ++it){
            auto itU = distinctDataEdges.find(it->second);
            if(itU == distinctDataEdges.end()){
                distinctDataEdges.insert(std::make_pair(it->second, 1));
//                edgeTrie->AddMultiedges(it->second); // add only when it is a new multiedge |
            }
            else
                ++itU->second;
        }

        std::set<std::set<int>> distinctEdges; // These edges are lexicographically ordered |
        CollectAllDataEdges(distinctDataEdges, distinctEdges);
//        CollectClosedDataEdges(distinctDataEdges, distinctEdges); // Use FP trie to retrieve only closed itemsets for edgeSet |

cout << "# distinct multiedge set: " << distinctDataEdges.size() << endl;
cout << "# distinct itemsets: " << distinctEdges.size() << endl;
cout << endl;

        /// Create structure for the edge subgraphSeeds
        for(auto it = distinctEdges.begin(); it != distinctEdges.end(); ++it){
            Subgraph subgraph;
            subgraph.edgeLabels.insert(std::make_pair(std::make_pair(0,1), *it)); // add the pattern with nodes 0 and 1 |
//            subgraph.frequency = it->second;
            subgraph.multiEdges.insert(*it);
            subgraphSeeds.push_back(subgraph);
        }
    }
}

void SubgraphMinerNauty::ExtractInfo(const Subgraph& subgraph, GraphParameter& subgraphInfo)
{
/// Node labels exist
/*
    Update: subgraphInfo.attributes
    */
/// ###

    std::set<int> nodes;
    for(auto it = subgraph.edgeLabels.begin(); it != subgraph.edgeLabels.end(); ++it){
        nodes.insert(it->first.first);
        nodes.insert(it->first.second);
    }
    subgraphInfo.nNodes = nodes.size();

    subgraphInfo.eLabelMap = subgraph.edgeLabels; // add the original edges |
    for(auto it = subgraph.edgeLabels.begin(); it != subgraph.edgeLabels.end(); ++it){ // add the reversed edges, for the undirected graphs |
        std::pair<int, int> edgeRev = std::make_pair(it->first.second, it->first.first);
        subgraphInfo.eLabelMap.insert(std::make_pair(edgeRev, it->second));
    }

    Vector2D adjListTemp(subgraphInfo.nNodes);
    EdgeLabel nbrSignTemp(subgraphInfo.nNodes);
    for(auto it = subgraph.edgeLabels.begin(); it != subgraph.edgeLabels.end(); ++it){
        nbrSignTemp[it->first.first].push_back(it->second);
        nbrSignTemp[it->first.second].push_back(it->second);
        if (find(adjListTemp[it->first.first].begin(), adjListTemp[it->first.first].end(), it->first.second) == adjListTemp[it->first.first].end())
          adjListTemp[it->first.first].push_back(it->first.second);
        if (find(adjListTemp[it->first.second].begin(), adjListTemp[it->first.second].end(), it->first.first) == adjListTemp[it->first.second].end())
          adjListTemp[it->first.second].push_back(it->first.first);
    }
    subgraphInfo.adjacencyList = adjListTemp;
    subgraphInfo.neighbourSign = nbrSignTemp;
}

void SubgraphMinerNauty::FindFrequentSeeds(const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newSubgraphs, std::deque<Subgraph>& newFSGs)
{
    for(size_t i = 0; i < newSubgraphs.size(); ++i){
        GraphParameter subgraphInfo;
        ExtractInfo(newSubgraphs[i], subgraphInfo);
        Sumgra subIsomorphism;


        subIsomorphism.FindEmbeddings(supportVal, subgraphInfo, dataGraphInfo, graphIndexes, newSubgraphs[i]);
        if(!newSubgraphs[i].embeddings.empty()){
//cout << "Initial Emb: " << newSubgraphs[i].embeddings.size() << endl;
            Frequency supportMeasure;
            if(supportMeasure.IsMinImageFrequent(newSubgraphs[i].embeddings, supportVal))
                newFSGs.push_back(newSubgraphs[i]);
        }


//        if(subIsomorphism.FindEmbeddings(supportVal, subgraphInfo, dataGraphInfo, graphIndexes, newSubgraphs[i]))
//            newFSGs.push_back(newSubgraphs[i]);

    }
}

/*
bool SubgraphMinerNauty::IsNotIsomorphicCurrentNauty(Subgraph& newSubgraph, std::vector<sparsegraph>& canonicalSet, const IsoTestMap& dataGraph)
{
    ExtractInfo(newSubgraph, newSubgraph.parameters);
    TracesIso subgraph;
//cout << "Can Gen begun" << endl;
    subgraph.GenerateCanonical(newSubgraph);
//cout << "Can Gen ended" << endl;
    for(size_t i = 0; i < canonicalSet.size(); ++i){
//        if(newSubgraph.multiEdges == )
        if (aresame_sg(&canonicalSet[i],&newSubgraph.canonicalNauty))
            return false; // the same subgraph already exists |
    }
    canonicalSet.push_back(newSubgraph.canonicalNauty);
    return true;
}
*/

/*
bool SubgraphMinerNauty::IsNotIsomorphicPreviousNauty(const int& subgraphSize, sparsegraph& canonical, IsoMeasure& canonicalAllLevel)
{
//    if(subgraphSize == 2)
//        return true; // No need to check |
//    else{
        auto it = canonicalAllLevel.find(subgraphSize);
        if(it == canonicalAllLevel.end()){ // Create a new level and add canonical elements to it |
            std::vector<sparsegraph> tmp;
//cout << "New canonical level: " << subgraphSize << endl;
//for(size_t i = 0; i < canonical.size(); ++i)
//cout << canonical[i] << ", ";
//cout << endl;
            tmp.push_back(canonical);
            canonicalAllLevel.insert(std::make_pair(subgraphSize, tmp));
            return true;
        }
        else{
            for(size_t i = 0; i < it->second.size(); ++i){
                if (aresame_sg(&it->second[i],&canonical))
                    return false; // the same subgraph already exists |
            }
            it->second.push_back(canonical);
            return true;
        }
//    }
}
*/

/// TO BE USED.

bool SubgraphMinerNauty::IsNotIsomorphicCurrentNauty(Subgraph& newSubgraph, std::vector<Canonical>& canonicalSet, const IsoTestMap& dataGraph)
{
    ExtractInfo(newSubgraph, newSubgraph.parameters);
    TracesIso subgraph;
//cout << "Can Gen begun" << endl;
    subgraph.GenerateCanonical(newSubgraph);
//cout << "Can Gen ended" << endl;
    for(size_t i = 0; i < canonicalSet.size(); ++i){
        if(newSubgraph.multiEdges == canonicalSet[i].multiEdges){ // Check for same edge types |
            if (aresame_sg(&canonicalSet[i].subgraph,&newSubgraph.canonicalNauty)) // Check for isomorphism |
                return false; // the same subgraph already exists |
        }
    }
    Canonical canonicalForm;
    canonicalForm.multiEdges = newSubgraph.multiEdges;
    canonicalForm.subgraph = newSubgraph.canonicalNauty;
    canonicalSet.push_back(canonicalForm);
    return true;
}

bool Testing(bool& val, const int& subgraphSize, Canonical& canonicalForm, IsoMeasure& canonicalAllLevel)
{

    auto it = canonicalAllLevel.find(subgraphSize);
    if(it == canonicalAllLevel.end()){ // Create a new level and add canonical elements to it |
//        Canonical canonicalForm; /// THIS FAULTY UNREMOVED LINE KILLED MY DAY. shucks.
        std::vector<Canonical> tmp;
        tmp.push_back(canonicalForm);
        canonicalAllLevel.insert(std::make_pair(subgraphSize, tmp));
        return true;
    }
    else{
        for(size_t i = 0; i < it->second.size(); ++i){
            if (it->second[i].multiEdges == canonicalForm.multiEdges){ // check for same edge types |
                if (aresame_sg(&it->second[i].subgraph,&canonicalForm.subgraph)){
if (subgraphSize == 5 && val)
cout << "EQUAL" << endl;
                    return false; // the same subgraph already exists |
                }
            }
        }
if (subgraphSize == 5 && val)
cout << "HERE????? => NOT EQUAL" << endl;
        it->second.push_back(canonicalForm);
        return true;
    }
}

bool SubgraphMinerNauty::IsNotIsomorphicPreviousNauty(const int& subgraphSize, Canonical& canonicalForm, IsoMeasure& canonicalAllLevel)
{

    auto it = canonicalAllLevel.find(subgraphSize);
    if(it == canonicalAllLevel.end()){ // Create a new level and add canonical elements to it |
//        Canonical canonicalForm; /// THIS FAULTY UNREMOVED LINE KILLED MY DAY. shucks.
        std::vector<Canonical> tmp;
        tmp.push_back(canonicalForm);
        canonicalAllLevel.insert(std::make_pair(subgraphSize, tmp));
        return true;
    }
    else{
//if(subgraphSize == 5)
//cout << "canonicalSet.size() " << it->second.size() << endl;
        for(size_t i = 0; i < it->second.size(); ++i){
            if (it->second[i].multiEdges == canonicalForm.multiEdges){ // check for same edge types |
/*
bool val = false;
if(canonicalForm.multiEdges.size() == 2 ){
    if (((*canonicalForm.multiEdges.begin()).size() == 1 && (*(*canonicalForm.multiEdges.begin()).begin()) == 1) && ((*canonicalForm.multiEdges.rbegin()).size() == 1 && (*(*canonicalForm.multiEdges.rbegin()).begin()) == 2))
        val = true;
}

bool val = true;
int k = 0;
for(auto it1 = canonicalForm.multiEdges.begin(); it1 != canonicalForm.multiEdges.end(); ++it1){
    if(k < 3){
        if((*it1).size() != 1 || *(*it1).begin() != 1){
            val = false;
            break;
        }
    }
    if(k > 2){
        if((*it1).size() != 1 || *(*it1).begin() != 2){
            val = false;
            break;
        }
    }
    ++k;
}
if(subgraphSize == 5 && val){
    for(auto it1 = canonicalForm.multiEdges.begin(); it1 != canonicalForm.multiEdges.end(); ++it1){
        for(auto it2 = (*it1).begin(); it2 != (*it1).end(); ++it2)
            cout << *it2 << ", ";
        cout << endl;
    }
    cout << "***" << endl;
}
*/
//cout << "NOT SAME " << endl;
                if (aresame_sg(&it->second[i].subgraph,&canonicalForm.subgraph))
                    return false; // the same subgraph already exists |
            }
        }
        it->second.push_back(canonicalForm);
        return true;
    }
}


inline bool NotAvalidExtension(const Set2D& nonFreqEdges, const std::set<int>& newEdge)
{
    for(auto it = nonFreqEdges.begin(); it != nonFreqEdges.end(); ++it)
        if(std::includes(newEdge.begin(), newEdge.end(), (*it).begin(), (*it).end()))
            return true;
    return false;
}

bool SubgraphMinerNauty::IsFrequent(const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, Subgraph& newSubgraph)
{
    Sumgra subIsomorphism;


    subIsomorphism.FindEmbeddings(supportVal, newSubgraph.parameters, dataGraphInfo, graphIndexes, newSubgraph);
    if(!newSubgraph.embeddings.empty()){
//cout << "Next Emb: " << newSubgraph.embeddings.size() << endl;
        Frequency supportMeasure;
        if(supportMeasure.IsMinImageFrequent(newSubgraph.embeddings, supportVal))
            return true;
    }
    return false;


//    if(subIsomorphism.FindEmbeddings(supportVal, newSubgraph.parameters, dataGraphInfo, graphIndexes, newSubgraph)){
////        cout << "returned sumgra" << endl;
//        return true;
//    }
//    else{
////        cout << "returned sumgra" << endl;
//        return false;
//    }
}


/// USING SET DATA STRUCTURE
void SubgraphMinerNauty::SubgraphExtension(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newFSGs)
{
/// Return only the frequent subgraphs.

//Ext 1: No of edges to be added = no. of nodes present in the subgraph. However, they contain isomorphic subgraphs |
//Ext 2: No of nodes to be added = Clique edges - existing edges in the subgraph. However, they may contain isomorphic subgraphs |
//cout << "subgraph Size: " << oldFSG.edgeLabels.size() << endl;
//    std::set<int> newEdge = fsgExtened.edgeLabels.begin()->second; // a new edge to be added |
    if(oldFSG.alreadyExtended){ // If the subgraph has already been extended |
//cout << "ALREADY EXTENDED" << endl;
        /// Ext 1. Adding an edge WITH an extra node.
/*
if(oldFSG.edgeLabels.size() == 5)
    cout << "REALLY?" << endl;
*/
        for(auto it = oldFSG.nonFreqEdgesN.begin(); it != oldFSG.nonFreqEdgesN.end(); ++it){
            Set2D nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for node *it
            if(!nonFreqEdges.empty()){ // Non-frequent edge extensions are available for this extension |
                if(NotAvalidExtension(nonFreqEdges, newEdge)){

//if(oldFSG.edgeLabels.size() == 2){
//cout << "subgraph Size: " << oldFSG.edgeLabels.size() << endl;
//for(auto itn = oldFSG.edgeLabels.begin()->second.begin(); itn != oldFSG.edgeLabels.begin()->second.end(); ++itn)
//cout << *itn << ", ";
//cout << endl;
//cout << it->first << ": ";
//for(auto itm = nonFreqEdges.begin(); itm != nonFreqEdges.end(); ++itm){
//for(auto itn = (*itm).begin(); itn != (*itm).end(); ++itn)
//cout << *itn << ", ";
//cout << ":: ";
//}
//cout << endl;
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//exit(0);
//}
                    continue; // try the next extension |

                }
            }
            std::pair<int, int> tmpEdgeID = std::make_pair(it->first, oldFSG.orderedNodes.size());
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
                edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge));
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |
/*
if(oldFSG.edgeLabels.size() == 5)
cout << *newEdge.begin() << endl;
*/
            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
            ExtractInfo(newSubgraph, newSubgraph.parameters);
            TracesIso subgraph;
            subgraph.GenerateCanonical(newSubgraph);
            Canonical canonicalForm;
            canonicalForm.multiEdges = multiEdges;
            canonicalForm.subgraph = newSubgraph.canonicalNauty;
            if(IsNotIsomorphicPreviousNauty(newSubgraph.edgeLabels.size(), canonicalForm, canonicalAllLevel)){
/*
if(oldFSG.edgeLabels.size() == 5)
cout << "NOT ISO!!!" << endl;
*/
                if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                    newFSGs.push_back(newSubgraph);
                }
                else{
                    it->second.insert(newEdge);
                }
            }
        }
        /// Ext 2. Adding an edge WITHOUT an extra node.
        for(auto it = oldFSG.nonFreqEdgesE.begin(); it != oldFSG.nonFreqEdgesE.end(); ++it){
            Set2D nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for edge *it
            if(!nonFreqEdges.empty()){ // Non-frequent edge extensions are available for this extension |
                if(NotAvalidExtension(nonFreqEdges, newEdge))
                    continue; // try the next extension |
            }
            std::pair<int,int> newEdgeIDRev = std::make_pair(it->first.second, it->first.first);
            if(oldFSG.edgeLabels.find(it->first) == oldFSG.edgeLabels.end() && oldFSG.edgeLabels.find(newEdgeIDRev) == oldFSG.edgeLabels.end()){
                EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
                edgeLabels.insert(std::make_pair(it->first, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
                Set2D multiEdges = oldFSG.multiEdges;
                multiEdges.insert(newEdge); // maintain the set of multiedges |

                Subgraph newSubgraph;
                newSubgraph.edgeLabels = edgeLabels;
                newSubgraph.multiEdges = multiEdges;
                ExtractInfo(newSubgraph, newSubgraph.parameters);
                TracesIso subgraph;
                subgraph.GenerateCanonical(newSubgraph);
                Canonical canonicalForm;
                canonicalForm.multiEdges = multiEdges;
                canonicalForm.subgraph = newSubgraph.canonicalNauty;
                if(IsNotIsomorphicPreviousNauty(newSubgraph.edgeLabels.size(), canonicalForm, canonicalAllLevel)){
                    if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                        newFSGs.push_back(newSubgraph);
                    }
                    else{
                        it->second.insert(newEdge);
                    }
                }
            }
        }
    }
    else{ // The subgraph "oldFSG" is being extended the first time |
        Set2D temp;
//if(oldFSG.edgeLabels.size() == 4){
//    cout << "FIRST TIME EXTENSION" << endl;
//for(auto it = oldFSG.edgeLabels.begin(); it != oldFSG.edgeLabels.end(); ++it){
//cout << it->first.first << ", " << it->first.second << ":: ";
//for(auto itn = it->second.begin(); itn != it->second.end(); ++itn)
//    cout << *itn << ",";
//cout << endl;
//}
//cout << endl;
////    exit(0);
//    }
        /// Ext 1. Adding an edge WITH an extra node.
        int nNodes = oldFSG.orderedNodes.size(); // no of nodes in the input subgraph |
        std::set<std::vector<int>> extraNodeCanonical;
        std::vector<Canonical> extraNodeCanonicalNauty;
        for(auto it = oldFSG.orderedNodes.begin(); it != oldFSG.orderedNodes.end(); ++it){
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            std::pair<int, int> tmpEdgeID = std::make_pair(*it, nNodes);
            if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
                edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge)); // add a new edge: a new node "nNodes" forming an edge with an already existing node "*it" |
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |

            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
            if(IsNotIsomorphicCurrentNauty(newSubgraph, extraNodeCanonicalNauty, dataGraph)){
                oldFSG.nonFreqEdgesN.insert(std::make_pair((*it), temp)); // This step assigns a set of unique nodes that are to be extended. Each node is selected on a first-come-first-serve basis, which is a representative node from its permutation group.
            Canonical canonicalForm;
            canonicalForm.multiEdges = multiEdges;
            canonicalForm.subgraph = newSubgraph.canonicalNauty;


bool val0 = true;
    int k = 0;
    for(auto it1 = newSubgraph.edgeLabels.begin(); it1 != newSubgraph.edgeLabels.end(); ++it1){
        if(k == 0 || k == 1 || k == 4){
            if(it1->second.size() != 1 || *it1->second.begin() !=1){
                val0 = false;
                break;
            }
        }
        if(k == 2 ||k == 3){
            if(it1->second.size() != 1 || *it1->second.begin() !=2){
                val0 = false;
                break;
            }
        }
        if((k == 4 && it1->first.first != 3) || (k == 3 && it1->first.first != 1) || (k == 2 && it1->first.first != 1)){
            val0 = false;
            break;
        }

        ++k;
    }
    if(newSubgraph.edgeLabels.size() == 5 && val0){
        cout << "##### EXTRA NODE OLD" << endl;
        for(auto it1 = newSubgraph.edgeLabels.begin(); it1 != newSubgraph.edgeLabels.end(); ++it1){
        cout << it1->first.first << ", " << it1->first.second << ":: ";
        for(auto itn = it1->second.begin(); itn != it1->second.end(); ++itn)
            cout << *itn << ",";
        cout << endl;
        }
        cout << "***" << endl;
    }

auto it0 = canonicalAllLevel.find(newSubgraph.edgeLabels.size());
bool val = true;
if(it0 != canonicalAllLevel.end()){
    int k = 0;
    for(auto it1 = newSubgraph.edgeLabels.begin(); it1 != newSubgraph.edgeLabels.end(); ++it1){
        if(k == 0 || k == 2 || k == 4){
            if(it1->second.size() != 1 || *it1->second.begin() !=1){
                val = false;
                break;
            }
        }
        if(k == 1 ||k == 3){
            if(it1->second.size() != 1 || *it1->second.begin() !=2){
                val = false;
                break;
            }
        }
        if((k == 4 && it1->first.first != 3) || (k == 3 && it1->first.first != 2) || (k == 2 && it1->first.first != 2)){
            val = false;
            break;
        }

        ++k;
    }
    if(newSubgraph.edgeLabels.size() == 5 && val){
        cout << "##### EXTRA NODE" << endl;
        for(auto it1 = newSubgraph.edgeLabels.begin(); it1 != newSubgraph.edgeLabels.end(); ++it1){
        cout << it1->first.first << ", " << it1->first.second << ":: ";
        for(auto itn = it1->second.begin(); itn != it1->second.end(); ++itn)
            cout << *itn << ",";
        cout << endl;
        }
        cout << "***" << endl;
    cout << "size: " << it0->second.size() << endl;

/*
Subgraph oldSubgraph;
EdgeLabelMap elabels;

std::set<int> s1;
s1.insert(1);
elabels.insert(std::make_pair(std::make_pair(0, 1), s1));

std::set<int> s2;
s2.insert(1);
elabels.insert(std::make_pair(std::make_pair(0, 2), s2));

std::set<int> s3;
s3.insert(2);
elabels.insert(std::make_pair(std::make_pair(1, 2), s3));

std::set<int> s4;
s4.insert(2);
elabels.insert(std::make_pair(std::make_pair(1, 3), s4));

std::set<int> s5;
s5.insert(1);
elabels.insert(std::make_pair(std::make_pair(3, 4), s5));

oldSubgraph.edgeLabels = elabels;
oldSubgraph.multiEdges = multiEdges;

ExtractInfo(oldSubgraph, oldSubgraph.parameters);
TracesIso sgOld;
sgOld.GenerateCanonical(oldSubgraph);
if (aresame_sg(&newSubgraph.canonicalNauty,&oldSubgraph.canonicalNauty))
    cout << "GREAT...THEY ARE SAME HERE TOO" << endl;
*/

    }
}

if(Testing(val, newSubgraph.edgeLabels.size(), canonicalForm, canonicalAllLevel)){


//            if(IsNotIsomorphicPreviousNauty(newSubgraph.edgeLabels.size(), canonicalForm, canonicalAllLevel)){


//                    oldFSG.nonFreqEdgesN.insert(std::make_pair((*it), temp));
                    if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                        newFSGs.push_back(newSubgraph);
                    }
                    else{
                        oldFSG.nonFreqEdgesN.find((*it))->second.insert(newEdge);


//cout << "subgraph Size: " << oldFSG.edgeLabels.size() << endl;
//cout << (*it) << ": ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl;
                    }
                }
            }
        }
        /// Ext 2. Adding an edge WITHOUT an extra node.
        std::set<std::pair<int,int>> extendableEdges;
        std::set<std::vector<int>> noNodeCanonical;
        std::vector<Canonical> noNodeCanonicalNauty;
        for(size_t i = 0; i < oldFSG.orderedNodes.size()-1; ++i){
            for(size_t j = i+1; j < oldFSG.orderedNodes.size(); ++j){
                std::pair<int,int> newEdgeID = std::make_pair(oldFSG.orderedNodes[i], oldFSG.orderedNodes[j]); // a possible edge in the subgraph clique |
                std::pair<int,int> newEdgeIDRev = std::make_pair(oldFSG.orderedNodes[j], oldFSG.orderedNodes[i]); // a possible edge in the subgraph clique |
                if(oldFSG.edgeLabels.find(newEdgeID) == oldFSG.edgeLabels.end() && oldFSG.edgeLabels.find(newEdgeIDRev) == oldFSG.edgeLabels.end()){
                    EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
                    edgeLabels.insert(std::make_pair(newEdgeID, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
                    Set2D multiEdges = oldFSG.multiEdges;
                    multiEdges.insert(newEdge); // maintain the set of multiedges |

                    Subgraph newSubgraph;
                    newSubgraph.edgeLabels = edgeLabels;
                    newSubgraph.multiEdges = multiEdges;
                    if(IsNotIsomorphicCurrentNauty(newSubgraph, noNodeCanonicalNauty, dataGraph)){
                        oldFSG.nonFreqEdgesE.insert(std::make_pair(newEdgeID, temp)); // This step assigns a set of unique node-pairs to which an edge can be added. Each node-pair is selected on a first-come-first-serve basis, which is a representative node-pair from its permutation group.
                        Canonical canonicalForm;
                        canonicalForm.multiEdges = multiEdges;
                        canonicalForm.subgraph = newSubgraph.canonicalNauty;



auto it0 = canonicalAllLevel.find(newSubgraph.edgeLabels.size());
bool val = true;
if(it0 != canonicalAllLevel.end()){
    int k = 0;
    for(auto it1 = newSubgraph.edgeLabels.begin(); it1 != newSubgraph.edgeLabels.end(); ++it1){
        if(k == 0 || k == 2 || k == 4){
            if(it1->second.size() != 1 || *it1->second.begin() !=1){
                val = false;
                break;
            }
        }
        if(k == 1 ||k == 3){
            if(it1->second.size() != 1 || *it1->second.begin() !=2){
                val = false;
                break;
            }
        }
        if((k == 4 && it1->first.first != 3) || (k == 3 && it1->first.first != 2) || (k == 2 && it1->first.first != 2)){
            val = false;
            break;
        }

        ++k;
    }
    if(newSubgraph.edgeLabels.size() == 5 && val){
        cout << "##### EXTRA NODE      NO" << endl;
        for(auto it1 = newSubgraph.edgeLabels.begin(); it1 != newSubgraph.edgeLabels.end(); ++it1){
        cout << it1->first.first << ", " << it1->first.second << ":: ";
        for(auto itn = it1->second.begin(); itn != it1->second.end(); ++itn)
            cout << *itn << ",";
        cout << endl;
        }
        cout << "***" << endl;
    cout << "size: " << it0->second.size() << endl;


//TracesIso sgNew;
//sgNew.GenerateCanonical(newSubgraph);


Subgraph oldSubgraph;
EdgeLabelMap elabels;

std::set<int> s1;
s1.insert(1);
elabels.insert(std::make_pair(std::make_pair(0, 1), s1));

std::set<int> s2;
s2.insert(1);
elabels.insert(std::make_pair(std::make_pair(0, 2), s2));

std::set<int> s3;
s3.insert(2);
elabels.insert(std::make_pair(std::make_pair(1, 2), s3));

std::set<int> s4;
s4.insert(2);
elabels.insert(std::make_pair(std::make_pair(1, 3), s4));

std::set<int> s5;
s5.insert(1);
elabels.insert(std::make_pair(std::make_pair(3, 4), s5));

oldSubgraph.edgeLabels = elabels;
oldSubgraph.multiEdges = multiEdges;
cout << "m Size " << multiEdges.size() << endl;
cout << "m Size " << newSubgraph.multiEdges.size() << endl;

ExtractInfo(oldSubgraph, oldSubgraph.parameters);
TracesIso sgOld;
sgOld.GenerateCanonical(oldSubgraph);
if (aresame_sg(&oldSubgraph.canonicalNauty,&newSubgraph.canonicalNauty))
    cout << "DAMN...THEY ARE SAME HERE" << endl;

    }
}

if(Testing(val, newSubgraph.edgeLabels.size(), canonicalForm, canonicalAllLevel)){


//                        if(IsNotIsomorphicPreviousNauty(newSubgraph.edgeLabels.size(), canonicalForm, canonicalAllLevel)){
                            if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                                newFSGs.push_back(newSubgraph);
                            }
                            else{
                                oldFSG.nonFreqEdgesE.find(newEdgeID)->second.insert(newEdge);
                            }
                        }
                    }
                }
            }
        }
        /// Comment the below line to UNPRUNE the unfrequent-edge-type-extension approach.
        oldFSG.alreadyExtended = true; // "oldFSG" has just been extended.
    }
}


/*
/// USING TRIE DATA STrUCTURE
void SubgraphMinerNauty::SubgraphExtension(const IsoTestMap& dataGraph, Subgraph& oldFSG, const set<int>& newEdge, IsoMeasure& canonicalAllLevel, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::deque<Subgraph>& newFSGs)
{
/// Return only the frequent subgraphs.
//std::vector<int> compare = {0,8,8};
//Ext 1: No of edges to be added = no. of nodes present in the subgraph. However, they contain isomorphic subgraphs |
//Ext 2: No of nodes to be added = Clique edges - existing edges in the subgraph. However, they may contain isomorphic subgraphs |

//    std::set<int> newEdge = fsgExtened.edgeLabels.begin()->second; // a new edge to be added |
//cout << "Extension started" << endl;
    if(oldFSG.alreadyExtended){ // If the subgraph has already been extended |
//cout << "already extended pattern" << endl;
        /// Ext 1. Adding an edge WITH an extra node.
        for(auto it = oldFSG.nonFreqEdgeN.begin(); it != oldFSG.nonFreqEdgeN.end(); ++it){
            Trie* nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for node *it
            if(nonFreqEdges->Root()->children().size() != 0){ // Non-frequent edge extensions are available for this extension |
                if(InvalidTrieEdge(nonFreqEdges, newEdge)){
//if(oldFSG.canonicalForm == compare){
//cout << "Preorder BEFORE: " << it->first << ": " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "New edge: ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//}
                    continue; // try the next extension |
                }
            }
            std::pair<int, int> tmpEdgeID = std::make_pair(it->first, oldFSG.orderedNodes.size());
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
                edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge));
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |
            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
            ExtractInfo(newSubgraph, newSubgraph.parameters);
            GraphIso subgraph;
            subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
            if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
                if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
                    newFSGs.push_back(newSubgraph);
//        cout << "returned sumgra 1" << endl;

                }
                else{
//        cout << "returned sumgra 1" << endl;

//if(oldFSG.canonicalForm == compare){
//cout << "Preorder BEFORE: " << it->first << ": " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//}
                    it->second->AddNonFreqEdges(newEdge);
//if(oldFSG.canonicalForm == compare){
//cout << "Preorder AFTER: " << it->first << ": " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//}
                }
            }
        }
        /// Ext 2. Adding an edge WITHOUT an extra node.
//if(compare == oldFSG.canonicalForm && newEdge.size() == 2){
//cout << "HERE : " << oldFSG.nonFreqEdgeE.size()  << endl;
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//Trie* nonFreqEdges = oldFSG.nonFreqEdgeE.begin()->second;
//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//}

        for(auto it = oldFSG.nonFreqEdgeE.begin(); it != oldFSG.nonFreqEdgeE.end(); ++it){
            Trie* nonFreqEdges = it->second; // "nonFreqEdges" specific to the possible extension for edge *it
//            if(nonFreqEdges->Root()->contentInt() == 0){ // Non-frequent edge extensions are available for this extension |
            if(nonFreqEdges->Root()->children().size() != 0){
//                if(NotAvalidExtension(nonFreqEdges, newEdge))
                if(InvalidTrieEdge(nonFreqEdges, newEdge)){
//if(compare == oldFSG.canonicalForm){
//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//cout << "New edge: ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn+1 << ",";
//cout << endl << "***" << endl;
//}

//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ",";
//cout << endl << "***" << endl;
//                    continue; // try the next extension |
                }
            }
            std::pair<int,int> newEdgeIDRev = std::make_pair(it->first.second, it->first.first);

            if(oldFSG.edgeLabels.find(it->first) == oldFSG.edgeLabels.end() && oldFSG.edgeLabels.find(newEdgeIDRev) == oldFSG.edgeLabels.end()){
                EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
                edgeLabels.insert(std::make_pair(it->first, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
                Set2D multiEdges = oldFSG.multiEdges;
                multiEdges.insert(newEdge); // maintain the set of multiedges |
                Subgraph newSubgraph;
                newSubgraph.edgeLabels = edgeLabels;
                newSubgraph.multiEdges = multiEdges;
                ExtractInfo(newSubgraph, newSubgraph.parameters);
                GraphIso subgraph;
                subgraph.GenerateCanonical(newSubgraph.parameters, newSubgraph.canonicalForm, dataGraph);
                if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
                    if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
//        cout << "returned sumgra 2" << endl;

                        newFSGs.push_back(newSubgraph);
//if(compare == oldFSG.canonicalForm)
//cout << "frequent " << newFSGs.size() << endl;

                    }
                    else{
//        cout << "returned sumgra 2" << endl;

//                        cout << "???" << endl;
                        it->second->AddNonFreqEdges(newEdge);
//if(compare == oldFSG.canonicalForm){
//Trie* nonFreqEdges = it->second;
//cout << "Preorder: " << endl;
//nonFreqEdges->preOrder(nonFreqEdges->Root());
//cout << "Postorder: " << endl;
//nonFreqEdges->postOrder(nonFreqEdges->Root());
//cout << "New edge: ";
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn+1 << ",";
//cout << endl << "***" << endl;
//}
                    }
                }
            }
        }
    }
    else{ // The subgraph "oldFSG" is being extended the first time |
//cout << "Initial extension" << endl;
//        newTrie->Root()->setContentInt(-1);
//if(oldFSG.edgeLabels.size() == 1 && oldFSG.edgeLabels.begin()->second.size() == 2)
//    cout << "HELLLOOOOOOO: " << *oldFSG.edgeLabels.begin()->second.begin() << ": " << *newEdge.begin() << endl;
        /// Ext 1. Adding an edge WITH an extra node.
        int nNodes = oldFSG.orderedNodes.size(); // no of nodes in the input subgraph |
        std::set<std::vector<int>> extraNodeCanonical;
        for(auto it = oldFSG.orderedNodes.begin(); it != oldFSG.orderedNodes.end(); ++it){
            EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
            std::pair<int, int> tmpEdgeID = std::make_pair(*it, nNodes);
            if(edgeLabels.find(tmpEdgeID) == edgeLabels.end())
                edgeLabels.insert(std::make_pair(tmpEdgeID, newEdge)); // add a new edge: a new node "nNodes" forming an edge with an already existing node "*it" |
            Set2D multiEdges = oldFSG.multiEdges;
            multiEdges.insert(newEdge); // maintain the set of multiedges |
            Subgraph newSubgraph;
            newSubgraph.edgeLabels = edgeLabels;
            newSubgraph.multiEdges = multiEdges;
//cout << "Check current ISO 3 " << edgeLabels.size() << endl;
            if(IsNotIsomorphicCurrent(newSubgraph, extraNodeCanonical, dataGraph)){ // Check if this new set of edges "edgeLabels" are isomorphic to the already existing "newSubgraphs"
//cout << " current ISO checked 3" << endl;

                Trie* newTrie = new Trie();
                oldFSG.nonFreqEdgeN[(*it)] = newTrie;
//cout << "Check previous ISO 3" << endl;
                if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//cout << "ISO checked 3" << endl;
                    if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
//        cout << "returned sumgra 3" << endl;

                        newFSGs.push_back(newSubgraph);
                    }
                    else{
//        cout << "returned sumgra 3" << endl;

                        oldFSG.nonFreqEdgeN.find((*it))->second->AddNonFreqEdges(newEdge);
                    }
                }
            }
        }
        /// Ext 2. Adding an edge WITHOUT an extra node.
        std::set<std::pair<int,int>> extendableEdges;
        std::set<std::vector<int>> noNodeCanonical;
        for(size_t i = 0; i < oldFSG.orderedNodes.size()-1; ++i){
            for(size_t j = i+1; j < oldFSG.orderedNodes.size(); ++j){
                std::pair<int,int> newEdgeID = std::make_pair(oldFSG.orderedNodes[i], oldFSG.orderedNodes[j]); // a possible edge in the subgraph clique |
                std::pair<int,int> newEdgeIDRev = std::make_pair(oldFSG.orderedNodes[j], oldFSG.orderedNodes[i]); // a possible edge in the subgraph clique |
                if(oldFSG.edgeLabels.find(newEdgeID) == oldFSG.edgeLabels.end() && oldFSG.edgeLabels.find(newEdgeIDRev) == oldFSG.edgeLabels.end()){
                    EdgeLabelMap edgeLabels = oldFSG.edgeLabels;
                    edgeLabels.insert(std::make_pair(newEdgeID, newEdge)); // add a new edge: // add the new edge only in one direction "newEdgeID" |
                    Set2D multiEdges = oldFSG.multiEdges;
                    multiEdges.insert(newEdge); // maintain the set of multiedges |
                    Subgraph newSubgraph;
                    newSubgraph.edgeLabels = edgeLabels;
                    newSubgraph.multiEdges = multiEdges;
//cout << "Check current ISO 4 " << edgeLabels.size() << endl;
                    if(IsNotIsomorphicCurrent(newSubgraph, noNodeCanonical, dataGraph)){ // Check if this new set of edges "edgeLabels" are isomorphic to the already existing "newSubgraphs"
//cout << " current ISO checked 4" << endl;

                        Trie* newTrie = new Trie();
                        oldFSG.nonFreqEdgeE[newEdgeID] = newTrie;//if(compare == oldFSG.canonicalForm)//if(compare == oldFSG.canonicalForm)
//cout << "Check previous ISO 4" << endl;
                        if(IsNotIsomorphicPrevious(newSubgraph.edgeLabels.size(), newSubgraph.canonicalForm, canonicalAllLevel)){
//cout << "ISO checked 4" << endl;

//                            oldFSG.nonFreqEdgeE[newEdgeID] = newTrie;
                            if(IsFrequent(supportVal, dataGraphInfo, graphIndexes, newSubgraph)){
//        cout << "returned sumgra 4 " << endl;

                                newFSGs.push_back(newSubgraph);
                            }
                            else{
//        cout << "returned sumgra 4"  << endl;

                                oldFSG.nonFreqEdgeE.find(newEdgeID)->second->AddNonFreqEdges(newEdge);
                            }
                        }
                    }
                }
            }
        }
        oldFSG.alreadyExtended = true; // "oldFSG" has just been extended.
    }
//    cout << "Exit EXtension" << endl;
}
*/

/*
void SubgraphMinerNauty::FindFrequentSubgraphs(TimeEval& timeEval, const std::set<int> newEdge, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, Subgraph& oldFSG, std::deque<Subgraph>& newSubgraphs, std::deque<Subgraph>& newFSGs)
{
    for(size_t i = 0; i < newSubgraphs.size(); ++i){
        Sumgra subIsomorphism;
            clock_t Time = clock();
//        cout << "hello 1" <<  endl;
//        cout << newSubgraphs[i].parameters.eLabelMap.size() << endl;
        subIsomorphism.FindEmbeddings(supportVal, newSubgraphs[i].parameters, dataGraphInfo, graphIndexes, newSubgraphs[i]);
//        cout << "hello 2" <<  endl;

            timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
        if(!newSubgraphs[i].embeddings.empty()){
            Frequency supportMeasure;
            if(supportMeasure.IsMinImageFrequent(newSubgraphs[i].embeddings, supportVal)){
                newFSGs.push_back(newSubgraphs[i]);
            }
            else{
                oldFSG.nonFreqEdgeExists = true;
                oldFSG.nonFreqEdges.insert(newEdge);
                oldFSG.nonFreqEdge = new Trie();
                oldFSG.nonFreqEdge->AddMultiedges(newEdge);
//for(auto it = newEdge.begin(); it != newEdge.end(); ++it)
//    cout << *it << ",";
//cout << endl;
            }
        }
        else{
            oldFSG.nonFreqEdgeExists = true;
            oldFSG.nonFreqEdges.insert(newEdge);
            oldFSG.nonFreqEdge = new Trie();
            oldFSG.nonFreqEdge->AddMultiedges(newEdge);
//for(auto it = newEdge.begin(); it != newEdge.end(); ++it)
//    cout << *it << ",";
//cout << endl;
        }
    }
}
*/

void SubgraphMinerNauty::CollectFSG(const int& sSize, const std::deque<Subgraph> newFSGs, std::map<int, std::deque<FSG>>& allFSG)
{
    std::deque<FSG> newFSG;
    for(size_t i = 0; i < newFSGs.size(); ++i){
        FSG iNewFSG;
        iNewFSG.structure = newFSGs[i].edgeLabels;
        newFSG.push_back(iNewFSG);
    }
    auto it = allFSG.find(sSize);
    if(it == allFSG.end())
        allFSG.insert(std::make_pair(sSize, newFSG));
    else
        it->second.insert(it->second.end(), newFSG.begin(), newFSG.end());
}

inline bool IsvalidEdge(const Subgraph& nextSubgraph, const std::set<int>& newEdge)
{
    for(auto it = nextSubgraph.nonFreqEdges.begin(); it != nextSubgraph.nonFreqEdges.end(); ++it)
        if(std::includes(newEdge.begin(), newEdge.end(), (*it).begin(), (*it).end()))
            return false;
    return true;
}


inline bool SubgraphMinerNauty::InvalidTrieEdge(Trie* trie, std::set<int> newEdge)
{
    std::set<int> newEdgeTemp = newEdge;
    /// The for loop is necessary, so as to accomplish the superset containment.
/// Work on "IfEdgeExists" function; it does not execute accurately yet.
    for(auto it = newEdge.begin(); it != newEdge.end(); ++it){
        bool found = false;
        int pos = 0;
//        if(trie->EdgeExists(trie->Root(), newEdgeTemp, pos, found))
//            return true;
        trie->EdgeExists(trie->Root(), newEdgeTemp, pos, found);
        if(found)
            return true;

        newEdgeTemp.erase(newEdgeTemp.begin());
    }
    return false;
}

void SubgraphMinerNauty::DfsRecursiveMining(StackFSG& fsgClassStack, IsoMeasure& canonicalAllLevel, std::deque<Subgraph>& extendableEdges, int nE, TimeEval& timeEval, const IsoTestMap& dataGraph, const int& supportVal, const GraphParameter& dataGraphInfo, IndexType& graphIndexes, std::map<int, std::deque<FSG>>& allFSG, int& noOfFsg)
{
//cout << "~~~ NE " << nE << endl;
    if (nE == extendableEdges.size())
        return; // go to next extension |
//cout << "nE: " << nE << endl;
//    std::deque<std::deque<Subgraph>> fsgClassStack; // initialize FSG-class stack for DFS tree |
//    fsgClassStack.push_front(initialFSG); // add the first element "size=1" to the stack |
//cout << "fsgClassStack - intial: " << fsgClassStack.size() << endl;


    while(!fsgClassStack.empty()){ // iterate until the entire DFS tree is spanned |
//        Subgraph oldFSG = fsgClassStack.front().front();
//cout << "oldFSG" << endl;
//for(auto itm = fsgClassStack.front().front().edgeLabels.begin(); itm != fsgClassStack.front().front().edgeLabels.end(); ++itm){
//cout << itm->first.first << ", " << itm->first.second << ": ";
//for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
//cout << *itn << ", ";
//cout << " - edge";
//}
//cout << endl;
        std::deque<Subgraph> newFSGs;
        std::set<int> newEdge = extendableEdges[nE].edgeLabels.begin()->second;
//cout << "newEdge: " ;
//for(auto itn = newEdge.begin(); itn != newEdge.end(); ++itn)
//cout << *itn << ", ";
//cout << endl;
//cout << "Wait? " << endl;
        SubgraphExtension(dataGraph, fsgClassStack.front().front(), newEdge, canonicalAllLevel, supportVal, dataGraphInfo, graphIndexes, newFSGs);
//cout << "NO Wait! " << endl;
        if (!newFSGs.empty()){ // Grow the DFS tree until we find frequent subgraphs ; else go back to the next stack element |
//cout << " >>>> Extended" << endl;

            fsgClassStack.push_front(newFSGs); // Add the new FSG to the stack |
//cout << "fsgClassStack - Next: " << fsgClassStack.size() << endl;
            CollectFSG(newFSGs.front().edgeLabels.size(), newFSGs, allFSG);
            noOfFsg += newFSGs.size(); // count all the subgraphs |
// Output FSG sizes in terms of edges and vertices |
//for(size_t i  = 0; i < newFSGs.size(); ++i)
//    cout << "Size: " << newFSGs[i].edgeLabels.size() << " Nodes: " << newFSGs[i].orderedNodes.size() << endl;


//cout << "Total FSG: " << noOfFsg << endl;
//cout << "newFSGs: " << newFSGs.size() << endl;
//for(size_t i  = 0; i < newFSGs.size(); ++i){
//for(auto itm = newFSGs[i].edgeLabels.begin(); itm != newFSGs[i].edgeLabels.end(); ++itm){
//cout << itm->first.first << ", " << itm->first.second << ": ";
//for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
//cout << *itn << ", ";
//cout << " Edge; ";
//}
//cout << endl;
//cout << "canonical Form:" << endl;
//for(size_t j = 0; j < newFSGs[i].canonicalForm.size(); ++j)
//cout << newFSGs[i].canonicalForm[j] << ", ";
//cout << endl;
//}
        }
        else{ // Adding "newEdge" to "oldFSG" did not produce newFSGs |
//cout << " >>>> Not Extended" << endl;
            DfsRecursiveMining(fsgClassStack, canonicalAllLevel, extendableEdges, nE+1, timeEval, dataGraph, supportVal, dataGraphInfo, graphIndexes, allFSG, noOfFsg);
//cout << "**************    RETURNED      ************ " << endl;
//cout << "----------------------------Recursive? " << nE << endl;
            /// Remove the FSG of maximal size in the stack, that could not be extended at all.
            if(!fsgClassStack.empty()){
//cout << "NEW stack FSG" << endl;
//for(auto itm = fsgClassStack.front().front().edgeLabels.begin(); itm != fsgClassStack.front().front().edgeLabels.end(); ++itm){
//cout << itm->first.first << ", " << itm->first.second << ": ";
//for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
//cout << *itn << ", ";
//cout << " - edge";
//}
//cout << endl;

                if(fsgClassStack.front().size() == 1)
                    fsgClassStack.pop_front(); // remove en entire "deque"
                else
                    fsgClassStack.front().pop_front(); // remove only the deque-element
            }
//cout << "fsgClassStack - LAST: " << fsgClassStack.size() << endl;
        }
        nE = 0; // reset the pointer for next Extension - nE |
    }
}


void SubgraphMinerNauty::MineFrequentSubgraphs(const GraphParameter& dataGraphInfo,  IndexType& graphIndexes, const int& supportVal, std::map<int, std::deque<FSG>>& allFSG)
{
    /// Create initial seeds for mining.
    std::deque<Subgraph> subgraphSeeds, fsgSeeds;
    CollectSubgraphSeeds(dataGraphInfo, subgraphSeeds);
    cout << "Subgraph seeds: " << subgraphSeeds.size() << endl;
    FindFrequentSeeds(supportVal, dataGraphInfo, graphIndexes, subgraphSeeds, fsgSeeds);
    cout << "Frequent seeds: " << fsgSeeds.size() << endl;

    /// Maintain a multiedge and vertex attribute mapping required for canonical representation of frequent subgraphs.
    IsoTestMap dataGraph;
    std::vector<int> edgeSize(fsgSeeds.size());
    for(size_t i = 0; i < fsgSeeds.size(); ++i){
        std::set<int> freqEdge = fsgSeeds[i].edgeLabels.begin()->second;
        dataGraph.multiedgeMap.insert(std::make_pair(freqEdge, i+1));
        edgeSize[i] = freqEdge.size();
    }
    std::vector<int> ind = SortIndexIncr(edgeSize);

    /// Collect seed FSGs
    CollectFSG(1, fsgSeeds, allFSG);
    std::deque<FSG> frequentSubgraphs; // for better ordering of patterns use above data structure |
    int noOfFsg = fsgSeeds.size(); // initialize the frequent subgraphs to the number of size-1 frequent subgraphs |
    TimeEval timeEval;

    std::deque<Subgraph> fsgSeedsSorted(fsgSeeds.size());
    for(size_t i = 0; i < fsgSeeds.size(); ++i)
        fsgSeedsSorted[i] = fsgSeeds[ind[i]];

    for(size_t i = 0; i < fsgSeedsSorted.size(); ++i){
        std::set<int> freqEdge = fsgSeedsSorted[i].edgeLabels.begin()->second;
        for(auto it = freqEdge.begin(); it != freqEdge.end(); ++it)
            cout << *it << ",";
        cout << endl;
    }

cout << "fsgSeedsSorted.size(): " << fsgSeedsSorted.size() << endl;
    for (size_t iS = 0; iS < fsgSeedsSorted.size(); ++iS) { // iS -> subgraph iterator |
//cout << iS << endl;
        // Add only the lexicographically GE (>=) elements |
        std::deque<Subgraph> extendableEdges(fsgSeedsSorted.begin()+iS, fsgSeedsSorted.end());
        int nE = 0; // initialize a pointer to the nextEdge to be added|
        IsoMeasure canonicalAllLevel; // For the DFS subtree that grows from "fsgSeedsSorted[iS]" |
        StackFSG fsgClassStack; // FSG-class stack for DFS tree |
        fsgClassStack.push_front(std::deque<Subgraph>(fsgSeedsSorted.begin()+iS, fsgSeedsSorted.begin()+iS+1));
        DfsRecursiveMining(fsgClassStack, canonicalAllLevel, extendableEdges, nE+0, timeEval, dataGraph, supportVal, dataGraphInfo, graphIndexes, allFSG, noOfFsg);
    }

    /// Output all FSGs to file.
/*
    std::ofstream oFile;
    oFile.open("/home/vijay/Phd/MULTIGRAPHS/trunk/dataset/mining/testM/res2.txt");
    for(auto it = allFSG.begin(); it != allFSG.end(); ++it){
        oFile << "Level " << it->first << ": has " << it->second.size() << " patterns." << endl;
        for(size_t i  = 0; i < it->second.size(); ++i){
//            oFile << "Pattern " << i << endl;
            for(auto itm = it->second[i].structure.begin(); itm != it->second[i].structure.end(); ++itm){
                oFile << itm->first.first << ", " << itm->first.second << ": ";
                for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
                    oFile << *itn << "";
                    oFile << " E;  ";
            }
            oFile << endl;
        }
        oFile << endl;
    }
    oFile.close();
*/
    /// Output all FSGs to screen.
/*
    for(auto it = allFSG.begin(); it != allFSG.end(); ++it){
        cout << "Level: " << it->first << " has " << it->second.size() << " patterns." << endl;
        for(size_t i  = 0; i < it->second.size(); ++i){
            cout << "Pattern " << i << endl;
            for(auto itm = it->second[i].structure.begin(); itm != it->second[i].structure.end(); ++itm){
                for(auto itn = itm->second.begin(); itn != itm->second.end(); ++itn)
                    cout << *itn << ", ";
                    cout << " Edge; ";
            }
            cout << endl;
        }
        cout << endl;
    }
*/

    cout << endl << "No Of All FSG: " << noOfFsg << endl;
    cout << "Extend time: " << timeEval.SgExt << endl;
    cout << "Freq Comput: " << timeEval.SgFrq << endl;
    cout << "Embd Comput: " << timeEval.EmbEv << endl;

}

/// 1 things
//1. Trie approach is still incomplete. Check if the nonFrequentEdgeTrie has any word, which is a subset of newEdge.
