#include "Sumgra.h"
//#include "Trie.h"
//#include "sumgra/Trie.h"

Sumgra::Sumgra()
{
    //ctor
}

Sumgra::~Sumgra()
{
    //dtor
}


void Sumgra::BuildIndexes(GraphParameter& dataGraphInfo, IndexType& indexType)
{
    std::vector<int> srt_d_nodes(dataGraphInfo.nNodes);
    Index dataIndex;
    dataIndex.SortSignature(dataGraphInfo.neighbourSign, srt_d_nodes, dataGraphInfo.nNodes); // sort the neighborhood signature with decreasing size of the number of neighbors |

    dataIndex.BuildAttHash(dataGraphInfo.attributes, indexType.attributeHash);
    dataIndex.BuildSynTrie(dataGraphInfo.neighbourSign, dataGraphInfo.nNodes, indexType.synopsesTrie);

//    EdgeLabelBit dataBitSet(dataGraphInfo.nNodes);
//    dataIndex.BuildBitSign(dataGraphInfo.neighbourSign, dataGraphInfo.nNodes, dataBitSet);

    indexType.neighborTrie.resize(dataGraphInfo.nNodes);
    dataIndex.BuildNeighTrie(dataGraphInfo.adjacencyList, dataGraphInfo.eLabelMap, dataGraphInfo.nNodes, indexType.neighborTrie);

}

bool Sumgra::FindIsoMatch(Subgraph& query, IndexType& indexType, Vector2D& solutions)
{
    const GraphParameter queryGraphInfo = query.parameters;
    Match subGraph;

    /// "orderedNodes"  WILL HAVE BEEN computed, when called from "IsPreviousNonFrequent".
    std::vector<int> queryOrdering = query.orderedNodes;

    /// "orderedNodes" MAY/MAY NOT HAVE BEEN computed, when called from "IsAlreadyFSG".
    if(queryOrdering.empty())
        subGraph.OrderVertices(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, queryOrdering);

    Index dataIndex;

    /// Fetch the vertex attribute solutions for all the query vertices to be used when necessary.

    /// With vertex labels.
    VecOfSet nodeMatches(queryGraphInfo.nNodes);

    /// Without vertex labels

    std::vector<int> initialMatches;
    std::vector<int> solution(queryGraphInfo.nNodes);
    int initialVertex = queryOrdering[0];
    if (!queryGraphInfo.neighbourSign[initialVertex].empty())
        dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[initialVertex], indexType.synopsesTrie, initialMatches);
        if (!initialMatches.empty()){
            if(subGraph.CheckSI(initialMatches, queryGraphInfo, queryOrdering, solution, nodeMatches, indexType.neighborTrie)){
                solutions.push_back(solution);
                return true;
            }
        }
    return false;
}


bool FindInitialMatches(const int& freq, IndexType& indexType, std::map<int,int>& new_old, Vector2D& initialMatchesAll, Vector2D& queryOrderPerms, Subgraph& matchedSubgraph, const EdgeLabel& neighbourSign)
{
    Index dataIndex;
    int nNodes = queryOrderPerms[0].size();
    for(size_t i = 0; i < nNodes; ++i){
        int initialVertex;
        if(new_old.empty())
            initialVertex = queryOrderPerms[i][0];
        else
            initialVertex = new_old.find(queryOrderPerms[i][0])->second;
        if (!neighbourSign[initialVertex].empty())
            dataIndex.QuerySynTrie(neighbourSign[initialVertex], indexType.synopsesTrie, initialMatchesAll[i]);
        if(initialMatchesAll[i].size() < freq)  // Check the size of initial matches BEFORE PRUNING |
            return false;
        else{
            auto it = matchedSubgraph.invalidCands.find(initialVertex); // check if the initial query vertex of i^th permutation  has any invalid candidates |
            if(it != matchedSubgraph.invalidCands.end()){
                    initialMatchesAll[i] = FindDifference(initialMatchesAll[i], it->second); // Prune the invalid candidates, learned from the previous levels |
                if(initialMatchesAll[i].size() < freq) // Check the size of initial matches AFTER PRUNING |
                    return false;
            }
        }
    }
    return true;
}

bool Sumgra::FindEmbeddings(const int& freq, const GraphParameter& queryGraphInfo, const GraphParameter& dataGraphInfo, IndexType& indexType, Subgraph& matchedSubgraph)
{
    Match subGraph;
//    subGraph.OrderVertices(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph.orderedNodes);
    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);

    Index dataIndex;

    /// Fetch the vertex attribute solutions for all the query vertices to be used when necessary.

    /// With vertex labels.
    VecOfSet nodeMatches(queryGraphInfo.nNodes);
/*
    dataIndex.QueryAttHash(queryGraphInfo.attributes, indexType.attributeHash, nodeMatches);
cout << orderedNodes.size() << ": " << orderedNodes[0] << "::" orderedNodes[0=1] << endl;
    std::vector<int> initialMatches, edgeMatches;
    int initialVertex = matchedSubgraph.orderedNodes[0];
    if (!queryGraphInfo.neighbourSign[initialVertex].empty())
        dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[initialVertex], indexType.synopsesTrie, edgeMatches);
cout << "edgeMatches: " << edgeMatches,size() << endl;
    if (!nodeMatches[initialVertex].empty()) {
        for (size_t i = 0; i < edgeMatches.size(); ++i){
            if (nodeMatches[initialVertex].find(edgeMatches[i]) != nodeMatches[initialVertex].end())
                initialMatches.push_back(edgeMatches[i]);
        }
    }
*/
    /// Without vertex labels

    std::vector<int> initialMatches;
    int initialVertex = matchedSubgraph.orderedNodes[0];
    if (!queryGraphInfo.neighbourSign[initialVertex].empty())
        dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[initialVertex], indexType.synopsesTrie, initialMatches);

    /// Find all the  embeddings bouded by MAX_EMB, and then discovered the frequent patterns.
//    if (!initialMatches.empty())
//        subGraph.FindMatches(initialMatches, queryGraphInfo, matchedSubgraph.orderedNodes, nodeMatches, indexType.neighborTrie, matchedSubgraph.embeddings);
//cout << "initialMatches: " << initialMatches.size() << endl;
    /// Find embeddings only until support value is met.
//clock_t Time = clock();
        if (!initialMatches.empty()){
            if(!matchedSubgraph.invalidCands.empty()){
double before = initialMatches.size();
//cout << initialVertex << " :: " << matchedSubgraph.invalidCands.begin()->first << endl;
clock_t Time = clock();

                if(matchedSubgraph.invalidCands.begin()->first == initialVertex)
                    initialMatches = FindDifference(initialMatches, matchedSubgraph.invalidCands.begin()->second); // Prune the invalid candidates, learned from the previous levels |

                if(initialMatches.size() < freq)
                    return false;
matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
//cout << before  << " :: " << initialMatches.size() << "; Diff = " << before - initialMatches.size() << endl;
//cout << "% Pruned: " << (before-initialMatches.size())*100/before << endl;
//cout << "SG Size: " << matchedSubgraph.edgeLabels.size() << endl;
            }
            if(subGraph.FindMatchesMNI(freq, initialMatches, queryGraphInfo, matchedSubgraph.orderedNodes, nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands)){
//            if(subGraph.FindMniFrequentMatches(freq, initialMatches, queryGraphInfo, matchedSubgraph.orderedNodes, nodeMatches, indexType.neighborTrie)){
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;

                return true;
            }
        }
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;

    return false;
}


/*
bool Sumgra::FindEmbeddingsMNI(const int& freq, const GraphParameter& queryGraphInfo, const GraphParameter& dataGraphInfo, IndexType& indexType, Subgraph& matchedSubgraph)
{
    /// Compute a set of query sequencing permutations that allows equal distribution of matched nodes for all query nodes. This helps to quickly achieve MNI support.

//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
//for(size_t i = 0; i < queryGraphInfo.adjacencyList.size(); ++i){
//    cout << i << " : ";
//    for(size_t j = 0; j < queryGraphInfo.adjacencyList[i].size(); ++j)
//        cout << queryGraphInfo.adjacencyList[i][j] << ", "; cout << endl;
//}
//}
    Match subGraph;
    Vector2D queryOrderPerms(queryGraphInfo.nNodes);
    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);

    for(int i = 0; i < queryGraphInfo.nNodes; ++i){
        subGraph.OrderVerticesAllMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, queryOrderPerms[i], i);
//        queryOrderPerms[i] = matchedSubgraph.orderedNodes;

if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
cout << "query ordering: " ;
for(size_t j = 0; j < queryOrderPerms.size(); ++j)
cout << queryOrderPerms[i][j] << ", " ; cout << endl;
}
    }



//    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);
//    matchedSubgraph.orderedNodes = queryOrderPerms[0]; // just to convey the query information for the pattern |


    Index dataIndex;

    /// Fetch the vertex attribute solutions for all the query vertices to be used when necessary.

    /// With vertex labels.
    VecOfSet nodeMatches(queryGraphInfo.nNodes);

//    dataIndex.QueryAttHash(queryGraphInfo.attributes, indexType.attributeHash, nodeMatches);
//cout << orderedNodes.size() << ": " << orderedNodes[0] << "::" orderedNodes[0=1] << endl;
//    std::vector<int> initialMatches, edgeMatches;
//    int initialVertex = matchedSubgraph.orderedNodes[0];
//    if (!queryGraphInfo.neighbourSign[initialVertex].empty())
//        dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[initialVertex], indexType.synopsesTrie, edgeMatches);
//cout << "edgeMatches: " << edgeMatches,size() << endl;
//    if (!nodeMatches[initialVertex].empty()) {
//        for (size_t i = 0; i < edgeMatches.size(); ++i){
//            if (nodeMatches[initialVertex].find(edgeMatches[i]) != nodeMatches[initialVertex].end())
//                initialMatches.push_back(edgeMatches[i]);
//        }
//    }


    /// Compute the initial candidate solutions for all the query vertices.
    Vector2D initialMatchesAll(queryGraphInfo.nNodes);
    std::vector<int> initialMatchSize(queryGraphInfo.nNodes);
    for(size_t i = 0; i < queryGraphInfo.nNodes; ++i){
        if (!queryGraphInfo.neighbourSign[queryOrderPerms[i][0]].empty())
            dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[queryOrderPerms[i][0]], indexType.synopsesTrie, initialMatchesAll[i]);
        if(initialMatchesAll[i].size() < freq)  // Check the size of initial matches BEFORE PRUNING |
            return false;
        else{
            auto it = matchedSubgraph.invalidCands.find(queryOrderPerms[i][0]); // check if the initial query vertex of i^th permutation  has any invalid candidates |
            if(it != matchedSubgraph.invalidCands.end()){
//cout << "Before:: " << it->first << ": " << initialMatchesAll[i].size() << endl;
                    initialMatchesAll[i] = FindDifference(initialMatchesAll[i], it->second); // Prune the invalid candidates, learned from the previous levels |
//cout << "After::: " << it->first << ": " << initialMatchesAll[i].size() << endl;
                if(initialMatchesAll[i].size() < freq) // Check the size of initial matches AFTER PRUNING |
                    return false;
            }
            initialMatchSize[i] = initialMatchesAll[i].size();
        }
if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1)
cout <<  "initialMatchSize: " << initialMatchSize[i] << endl;
    }

    /// Find embeddings
    std::vector<std::unordered_set<int>> nodeRepository(queryGraphInfo.nNodes); // Maintains distinct node embeddings for each query node.
    std::vector<int> updatedMatchSize = initialMatchSize;
    int allNodesTraversed = 0;
    int j = 0;
    int operations = 0;
    while(allNodesTraversed < queryGraphInfo.nNodes){
//cout << allNodesTraversed << ", " << queryGraphInfo.nNodes << endl;
        for(size_t i = 0; i < queryOrderPerms.size(); ++i){
//if (i ==0 )
//cout << j << ", " << initialMatchSize[i] << endl;
int tmp = *queryOrderPerms[i].begin();
//if(!nodeRepository[tmp].empty())
//cout << nodeRepository[*queryOrderPerms[i].begin()].size() << ", " << nodeRepository[*queryOrderPerms[i].rbegin()].size() << endl;
if(nodeRepository[tmp].size() >= freq)
//    if(nodeRepository[queryGraphInfo.nNodes-1].size() >= freq && tmp == queryGraphInfo.nNodes-1)
        continue;
//    else
//        cout << "no" << endl;

            if(j < initialMatchSize[queryOrderPerms[i][0]]){ // make sure that you do not go out of memory bound.
                ++operations;
if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
cout << "queryOrderPerms: " ;
for(size_t j = 0; j < queryOrderPerms.size(); ++j)
cout <<  queryOrderPerms[i][j] << ", " ; cout << endl;


for(size_t i = 0; i < queryGraphInfo.nNodes; ++i)
cout << nodeRepository[i].size() << ", "; cout << endl;
}

                if(subGraph.FindMatchesSingleMNI(freq, initialMatchesAll[queryOrderPerms[i][0]][j], queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, updatedMatchSize[queryOrderPerms[i][0]], matchedSubgraph.invalidCands, nodeRepository)){
//if(initialMatchesAll.size() == 3)
//cout << "frequent" << endl;
cout << "# operations to succeed: " << operations << endl;
                    return true; // MNI support is reached |

                }
                else
                    if(updatedMatchSize[queryOrderPerms[i][0]] < freq)
                        return false; // initial match size shrunk below the support value |

            }
            else
                ++allNodesTraversed;
        }
        ++j;
    }
    return false;
}
*/

/// Outputs only one embedding per iteration.

bool Sumgra::FindEmbeddingsMNI(const int& freq, GraphParameter& queryGraphInfo, const GraphParameter& dataGraphInfo, IndexType& indexType, Subgraph& matchedSubgraph)
{
    int nCoreNodes = 0;
    for (size_t m = 0; m < queryGraphInfo.adjacencyList.size(); ++m){
        if (queryGraphInfo.adjacencyList[m].size() > 1)
            ++nCoreNodes ;
    }

    Match subGraph;
    Index dataIndex;

    /// With vertex labels.
    VecOfSet nodeMatches(queryGraphInfo.nNodes);

    std::vector<std::unordered_set<int>> nodeRepository(queryGraphInfo.nNodes); // Maintains distinct node embeddings for each query node.
    int requiredEmbs = 0;


    /// Case 1: If the pattern is a pure star.
//    if(nCoreNodes == 0 || nCoreNodes == 1){
    if(false){
//clock_t Time = clock();
        std::vector<int> orderedQuery;
        subGraph.OrderStarNodes(queryGraphInfo.neighbourSign, orderedQuery); // compute query ordering for a star pattern
        matchedSubgraph.orderedNodes = orderedQuery;

        std::vector<int> initialMatches;
        if (!queryGraphInfo.neighbourSign[orderedQuery[0]].empty())
            dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[orderedQuery[0]], indexType.synopsesTrie, initialMatches);

        auto it = matchedSubgraph.invalidCands.find(orderedQuery[0]);
        if(it != matchedSubgraph.invalidCands.end())
            initialMatches = FindDifference(initialMatches, it->second);

        if(initialMatches.size() < freq)
            return false;

        if(subGraph.FindStarMatches(freq, initialMatches, queryGraphInfo, orderedQuery, nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands, nodeRepository)){

//matchedSubgraph.timeEval.time1 += double(clock()-Time)/CLOCKS_PER_SEC;
            return true;
        }
//matchedSubgraph.timeEval.time1 += double(clock()-Time)/CLOCKS_PER_SEC;
        return false;
    }

    /// Compute a set of query sequencing permutations that allows equal distribution of matched nodes for all query nodes. This helps to quickly achieve MNI support.

//clock_t Time = clock();

    Vector2D queryOrderPerms(queryGraphInfo.nNodes);
    for(int i = 0; i < queryGraphInfo.nNodes; ++i)
        subGraph.OrderVerticesAllMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, queryOrderPerms[i], i);
    matchedSubgraph.orderedNodes = queryOrderPerms[0]; // A pattern needs to a have one permutation of query ordering

    /// Compute the initial candidate solutions for all the query vertices.
    Vector2D initialMatchesAll(queryGraphInfo.nNodes);
    if(!FindInitialMatches(freq, indexType, queryGraphInfo.new_old, initialMatchesAll, queryOrderPerms, matchedSubgraph, queryGraphInfo.neighbourSign)){
        return false;
    }

    std::vector<int> imSize(initialMatchesAll.size());
    for(size_t i = 0; i < initialMatchesAll.size(); ++i)
        imSize[i] = initialMatchesAll[i].size();

    std::vector<int> pO = SortIndexIncr(imSize); // order on the choice of permutation - "queryOrderPerms" |

    /// Case 2: If the pattern is a complex pattern.
/*
    /// Collect the satellite nodes, in the pattern.
    std::map<int, std::set<int>> satNodes; // <core node, satellite nodes> |
    std::set<std::pair<int,int>> satEdges; // <core-satellite> : edges corresponding to satellite nodes and core nodes |
    subGraph.GetSatelliteInfo(queryGraphInfo.adjacencyList, satNodes, satEdges);

//matchedSubgraph.timeEval.time2 += double(clock()-Time)/CLOCKS_PER_SEC;


//    if(!satNodes.empty()){
    if(false){
        /// Collect only the core nodes.
        std::unordered_set<int> coreNodes;
        for(auto it = satNodes.begin(); it != satNodes.end(); ++it)
            coreNodes.insert(it->first);

        /// Decompose the complex pattern into core nodes and satellite nodes.
        GraphParameter truncPattern;
        subGraph.DecomposePattern(satEdges, queryGraphInfo, truncPattern);

        /// Order satellite nodes of each core node.
        std::map<int, std::vector<int>> orderedSatNodes; // <core node, ordered satellite nodes>. This maintains the original vertex ids of the pattern |
        subGraph.OrderSatelliteNodes(queryGraphInfo, satNodes, orderedSatNodes);

        /// Order core nodes.
        Vector2D orderedCoreNodePerms(truncPattern.nNodes), orderedCoreOldPerms(truncPattern.nNodes);
        for(int i = 0; i < truncPattern.nNodes; ++i){
            subGraph.OrderVerticesAllMNI(truncPattern.nNodes, truncPattern.neighbourSign, truncPattern.adjacencyList, orderedCoreNodePerms[i], i);
            std::vector<int> temp(orderedCoreNodePerms[i].size());
            for(size_t j = 0; j < orderedCoreNodePerms[i].size(); ++j)
                temp[j] = truncPattern.new_old.find(orderedCoreNodePerms[i][j])->second;
            orderedCoreOldPerms[i] = temp;
        }

        /// 1. Compute pattern embeddings, by exploiting the query decomposition
        bool notFrequent = false;

            //1.For each query nodes, we collect satellite solutions as a vector multiplication instead of DFS.


        for(size_t i = 0; i < truncPattern.nNodes; ++i){
            int oldV = orderedCoreOldPerms[i][0];
            if(!nodeRepository[oldV].empty())
                initialMatchesAll[oldV] = FindDifferenceSet(initialMatchesAll[oldV], nodeRepository[oldV]);
            if(subGraph.FindComplexMatches(freq, initialMatchesAll[oldV], queryGraphInfo, truncPattern, satNodes, orderedSatNodes, orderedCoreNodePerms[i], orderedCoreOldPerms[i], nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands, nodeRepository, notFrequent)){
//cout << "NEW PATTERN F" << endl;
//for(size_t j = 0; j < initialMatchesAll.size(); ++j)
//    cout << initialMatchesAll[j].size() << endl;
//for(auto it = matchedSubgraph.edgeLabels.begin(); it != matchedSubgraph.edgeLabels.end(); ++it){
//    cout << it->first.first << ", " << it->first.second << ":: ";
//    for(auto itn = it->second.begin(); itn != it->second.end(); ++itn)
//        cout << *itn << ",";
//    cout << endl;
//}
//exit(0);
                return true;
            }
            else if(notFrequent){
//cout << "NEW PATTERN NOT F" << endl;
//for(size_t j = 0; j < initialMatchesAll.size(); ++j)
//    cout << initialMatchesAll[j].size() << endl;
//for(auto it = matchedSubgraph.edgeLabels.begin(); it != matchedSubgraph.edgeLabels.end(); ++it){
//    cout << it->first.first << ", " << it->first.second << ":: ";
//    for(auto itn = it->second.begin(); itn != it->second.end(); ++itn)
//        cout << *itn << ",";
//    cout << endl;
//}
//cout << endl;
//for(size_t p = 0; p < nodeRepository.size(); ++p)
//    cout << nodeRepository[p].size() << ", ";
//cout << endl;
//exit(0);
                return false;
            }
        }



//1. Following modules 2 and 3 can be executed interchangeably. The performance of the code depends on this order and is specific to a dataset.Choose the model-order that suits the dataset.
//2. If we do not execute model 2 (OPTION APPROXIMATE), our approach would be approximate; however, it is observed that in all the datasets, this approach has an approximate factor of 1. !!


        /// 3. Compute pattern embeddings, only for the core vertices as the initial vertex.

        for(auto it = coreNodes.begin(); it != coreNodes.end(); ++it){
            if(!nodeRepository[queryOrderPerms[*it][0]].empty())
                initialMatchesAll[queryOrderPerms[*it][0]] = FindDifferenceSet(initialMatchesAll[queryOrderPerms[*it][0]], nodeRepository[queryOrderPerms[*it][0]]);

            if(subGraph.FindMatchesFastMNI(freq, initialMatchesAll[queryOrderPerms[*it][0]], queryGraphInfo, queryOrderPerms[*it], nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands, nodeRepository, notFrequent)){
//        cout << "# Embeddings    (Success): " << requiredEmbs << endl;
                return true;
            }
            else if(notFrequent)
                return false;
        }


        /// 2. ***** OPTION APPROXIMATE ****** . Compute pattern embeddings, for the rest of the vertices (satellite vertices) as the initial vertex.

//            1. If we don't use this module, it will be an approximate version, with an approximation actor of 1 !!!. So check this scenario in all the test datasets.
//            2. Including this will give us the exact approach.


        for(size_t i = 0; i < queryGraphInfo.nNodes; ++i){
            if(coreNodes.find(i) != coreNodes.end())
                continue; // skip considering core vertices |
            if(!nodeRepository[queryOrderPerms[i][0]].empty())
                initialMatchesAll[queryOrderPerms[i][0]] = FindDifferenceSet(initialMatchesAll[queryOrderPerms[i][0]], nodeRepository[queryOrderPerms[i][0]]);
            if(subGraph.FindMatchesFastMNI(freq, initialMatchesAll[queryOrderPerms[i][0]], queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands, nodeRepository, notFrequent)){
                return true;
            }
            else if(notFrequent)
                return false;
        }
        return false;
    }
*/

    /// Case 3: If the pattern is a quasi-clique pattern
//    clock_t sTime = clock();
    bool notFrequent = false;
    for(size_t k = 0; k < queryGraphInfo.nNodes; ++k){
        /// You replace "queryOrderPerms[i][0]" by just "i", only if "queryOrderPerms" have contagious values from 0-n.
//clock_t Time = clock();
        int i = queryOrderPerms[pO[k]][0];
        if(!nodeRepository[i].empty())
            initialMatchesAll[i] = FindDifferenceSet(initialMatchesAll[i], nodeRepository[i]);


//matchedSubgraph.timeEval.time1 += double(clock()-Time)/CLOCKS_PER_SEC;

        if(subGraph.FindMatchesFastMNI(freq, initialMatchesAll[i], queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands, nodeRepository, notFrequent)){
//matchedSubgraph.timeEval.time2 += double(clock()-sTime)/CLOCKS_PER_SEC;
//cout << "Frequent" << endl;
//for(auto it = matchedSubgraph.edgeLabels.begin(); it != matchedSubgraph.edgeLabels.end(); ++it){
//    cout << it->first.first << ", " << it->first.second << ":: ";
//    for(auto itn = it->second.begin(); itn != it->second.end(); ++itn)
//        cout << *itn << ",";
//    cout << endl;
//}
//cout << endl;
            return true;
        }
        else if(notFrequent){
//matchedSubgraph.timeEval.time3 += double(clock()-sTime)/CLOCKS_PER_SEC;
            return false;
        }
    }
//matchedSubgraph.timeEval.time3 += double(clock()-sTime)/CLOCKS_PER_SEC;
//cout << "NON Frequent" << endl;
//for(auto it = matchedSubgraph.edgeLabels.begin(); it != matchedSubgraph.edgeLabels.end(); ++it){
//    cout << it->first.first << ", " << it->first.second << ":: ";
//    for(auto itn = it->second.begin(); itn != it->second.end(); ++itn)
//        cout << *itn << ",";
//    cout << endl;
//}
//cout << endl;

    return false;
}



/// EXPLOITS entire DFS tree path to outputs solutions.
/*
bool Sumgra::FindEmbeddingsMNI(const int& freq, const GraphParameter& queryGraphInfo, const GraphParameter& dataGraphInfo, IndexType& indexType, Subgraph& matchedSubgraph)
{
    /// Compute a set of query sequencing permutations that allows equal distribution of matched nodes for all query nodes. This helps to quickly achieve MNI support.

//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
//for(size_t i = 0; i < queryGraphInfo.adjacencyList.size(); ++i){
//    cout << i << " : ";
//    for(size_t j = 0; j < queryGraphInfo.adjacencyList[i].size(); ++j)
//        cout << queryGraphInfo.adjacencyList[i][j] << ", "; cout << endl;
//}
//}
    Match subGraph;
    Vector2D queryOrderPerms(queryGraphInfo.nNodes);
    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);

    for(int i = 0; i < queryGraphInfo.nNodes; ++i){
        subGraph.OrderVerticesAllMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, queryOrderPerms[i], i);
//        queryOrderPerms[i] = matchedSubgraph.orderedNodes;

//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
//cout << "query ordering: " ;
//for(size_t j = 0; j < queryOrderPerms.size(); ++j)
//cout << queryOrderPerms[i][j] << ", " ; cout << endl;
//}
    }



//    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);
//    matchedSubgraph.orderedNodes = queryOrderPerms[0]; // just to convey the query information for the pattern |


    Index dataIndex;

    /// Fetch the vertex attribute solutions for all the query vertices to be used when necessary.

    /// With vertex labels.
    VecOfSet nodeMatches(queryGraphInfo.nNodes);

//    dataIndex.QueryAttHash(queryGraphInfo.attributes, indexType.attributeHash, nodeMatches);
//cout << orderedNodes.size() << ": " << orderedNodes[0] << "::" orderedNodes[0=1] << endl;
//    std::vector<int> initialMatches, edgeMatches;
//    int initialVertex = matchedSubgraph.orderedNodes[0];
//    if (!queryGraphInfo.neighbourSign[initialVertex].empty())
//        dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[initialVertex], indexType.synopsesTrie, edgeMatches);
//cout << "edgeMatches: " << edgeMatches,size() << endl;
//    if (!nodeMatches[initialVertex].empty()) {
//        for (size_t i = 0; i < edgeMatches.size(); ++i){
//            if (nodeMatches[initialVertex].find(edgeMatches[i]) != nodeMatches[initialVertex].end())
//                initialMatches.push_back(edgeMatches[i]);
//        }
//    }


    /// Compute the initial candidate solutions for all the query vertices.
    Vector2D initialMatchesAll(queryGraphInfo.nNodes);
    std::vector<int> initialMatchSize(queryGraphInfo.nNodes);
    for(size_t i = 0; i < queryGraphInfo.nNodes; ++i){
        if (!queryGraphInfo.neighbourSign[queryOrderPerms[i][0]].empty())
            dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[queryOrderPerms[i][0]], indexType.synopsesTrie, initialMatchesAll[i]);
        if(initialMatchesAll[i].size() < freq)  // Check the size of initial matches BEFORE PRUNING |
            return false;
        else{
            auto it = matchedSubgraph.invalidCands.find(queryOrderPerms[i][0]); // check if the initial query vertex of i^th permutation  has any invalid candidates |
            if(it != matchedSubgraph.invalidCands.end()){
//cout << "Before:: " << it->first << ": " << initialMatchesAll[i].size() << endl;
                    initialMatchesAll[i] = FindDifference(initialMatchesAll[i], it->second); // Prune the invalid candidates, learned from the previous levels |
//cout << "After::: " << it->first << ": " << initialMatchesAll[i].size() << endl;
                if(initialMatchesAll[i].size() < freq) // Check the size of initial matches AFTER PRUNING |
                    return false;
            }
            initialMatchSize[i] = initialMatchesAll[i].size();
        }
    }

    /// Find embeddings. DFS way.
    std::vector<std::unordered_set<int>> nodeRepository(queryGraphInfo.nNodes); // Maintains distinct node embeddings for each query node.
    std::vector<int> updatedMatchSize = initialMatchSize;

    bool stackNotEmpty = true;
int j = 0, requiredEmbs = 0;
    while(stackNotEmpty){
        std::set<int> supportReached;
        for(size_t i = 0; i < nodeRepository.size(); ++i)
            if (nodeRepository[i].size() >= freq)
                supportReached.insert(i);


//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1)
//cout << "supportReached:                              " << supportReached.size() << endl;
        int stackMatchCount = 0;
        for(size_t i = 0; i < queryOrderPerms.size(); ++i){
            if(!supportReached.empty())
                if(supportReached.find(*queryOrderPerms[i].rbegin()) == supportReached.end()){
                    ++stackMatchCount;
                    continue;
                }

            int seedVertex;
            if(!initialMatchesAll[queryOrderPerms[i][0]].empty()){ // if initial vertex stack is not yet empty |
                seedVertex = initialMatchesAll[queryOrderPerms[i][0]].back();
                initialMatchesAll[queryOrderPerms[i][0]].pop_back();
            }
            else{
                ++stackMatchCount;
                continue;
            }
            bool matchSolutions = false;
            while(!matchSolutions){
                if(nodeRepository[queryOrderPerms[i][0]].find(seedVertex) == nodeRepository[queryOrderPerms[i][0]].end()){ // if the next initial seed vertex is never matched before |
                    matchSolutions = true; // an attempt is being made to find/match solutions |
//clock_t Time = clock();
                    if(subGraph.FindMatchesSingleMNI(freq, seedVertex, queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, updatedMatchSize[queryOrderPerms[i][0]], matchedSubgraph.invalidCands, nodeRepository, requiredEmbs)){
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
cout << "# Embeddings    (Success): " << requiredEmbs << endl;
                        return true;
                    }
                    else{
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
//cout << "# Embeddings (Failure): " << requiredEmbs << endl;

                        if(updatedMatchSize[queryOrderPerms[i][0]] < freq)
                            return false; // initial match size shrunk below the support value |
                    }
                }
                else{
                    if(!initialMatchesAll[queryOrderPerms[i][0]].empty()){ // if initial vertex stack is not yet empty |
                        seedVertex = initialMatchesAll[queryOrderPerms[i][0]].back();
                        initialMatchesAll[queryOrderPerms[i][0]].pop_back();
                    }
                    else{
                        ++stackMatchCount;
                        break;
                    }
                }
            }


//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
//cout << "Ordering: " ;
//for(size_t j = 0; j < queryOrderPerms.size(); ++j)
//cout << queryOrderPerms[i][j] << ", " ; cout << endl;
//}
        }

        if(stackMatchCount == queryOrderPerms.size()) // each of the stack bin has gone empty.
            stackNotEmpty = false;


//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
//cout  << "j=" << j << "::  ";
//for(size_t i = 0; i < queryOrderPerms.size(); ++i)
//    cout << initialMatchesAll[i].size() << ", "; cout << endl;
//cout << "\t ******" << endl;
//}

++j;
    }
    return false;
}
*/

/// EXPLOITS entire DFS tree path + single embedding per search path, to outputs solutions.
/*
bool Sumgra::FindEmbeddingsMNI(const int& freq, const GraphParameter& queryGraphInfo, const GraphParameter& dataGraphInfo, IndexType& indexType, Subgraph& matchedSubgraph)
{
    /// Compute a set of query sequencing permutations that allows equal distribution of matched nodes for all query nodes. This helps to quickly achieve MNI support.

//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
//for(size_t i = 0; i < queryGraphInfo.adjacencyList.size(); ++i){
//    cout << i << " : ";
//    for(size_t j = 0; j < queryGraphInfo.adjacencyList[i].size(); ++j)
//        cout << queryGraphInfo.adjacencyList[i][j] << ", "; cout << endl;
//}
//}
    Match subGraph;
    Vector2D queryOrderPerms(queryGraphInfo.nNodes);
    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);

    for(int i = 0; i < queryGraphInfo.nNodes; ++i){
        subGraph.OrderVerticesAllMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, queryOrderPerms[i], i);
//        queryOrderPerms[i] = matchedSubgraph.orderedNodes;

if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1){
cout << "query ordering: " ;
for(size_t j = 0; j < queryOrderPerms.size(); ++j)
cout << queryOrderPerms[i][j] << ", " ; cout << endl;
}
    }



//    subGraph.OrderVerticesMNI(queryGraphInfo.nNodes, queryGraphInfo.neighbourSign, queryGraphInfo.adjacencyList, matchedSubgraph);
//    matchedSubgraph.orderedNodes = queryOrderPerms[0]; // just to convey the query information for the pattern |


    Index dataIndex;

    /// Fetch the vertex attribute solutions for all the query vertices to be used when necessary.

    /// With vertex labels.
    VecOfSet nodeMatches(queryGraphInfo.nNodes);

//    dataIndex.QueryAttHash(queryGraphInfo.attributes, indexType.attributeHash, nodeMatches);
//cout << orderedNodes.size() << ": " << orderedNodes[0] << "::" orderedNodes[0=1] << endl;
//    std::vector<int> initialMatches, edgeMatches;
//    int initialVertex = matchedSubgraph.orderedNodes[0];
//    if (!queryGraphInfo.neighbourSign[initialVertex].empty())
//        dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[initialVertex], indexType.synopsesTrie, edgeMatches);
//cout << "edgeMatches: " << edgeMatches,size() << endl;
//    if (!nodeMatches[initialVertex].empty()) {
//        for (size_t i = 0; i < edgeMatches.size(); ++i){
//            if (nodeMatches[initialVertex].find(edgeMatches[i]) != nodeMatches[initialVertex].end())
//                initialMatches.push_back(edgeMatches[i]);
//        }
//    }


    /// Compute the initial candidate solutions for all the query vertices.
    Vector2D initialMatchesAll(queryGraphInfo.nNodes);
    std::vector<int> initialMatchSize(queryGraphInfo.nNodes);
    for(size_t i = 0; i < queryGraphInfo.nNodes; ++i){
        if (!queryGraphInfo.neighbourSign[queryOrderPerms[i][0]].empty())
            dataIndex.QuerySynTrie(queryGraphInfo.neighbourSign[queryOrderPerms[i][0]], indexType.synopsesTrie, initialMatchesAll[i]);
        if(initialMatchesAll[i].size() < freq)  // Check the size of initial matches BEFORE PRUNING |
            return false;
        else{
            auto it = matchedSubgraph.invalidCands.find(queryOrderPerms[i][0]); // check if the initial query vertex of i^th permutation  has any invalid candidates |
            if(it != matchedSubgraph.invalidCands.end()){
//cout << "Before:: " << it->first << ": " << initialMatchesAll[i].size() << endl;
                    initialMatchesAll[i] = FindDifference(initialMatchesAll[i], it->second); // Prune the invalid candidates, learned from the previous levels |
//cout << "After::: " << it->first << ": " << initialMatchesAll[i].size() << endl;
                if(initialMatchesAll[i].size() < freq) // Check the size of initial matches AFTER PRUNING |
                    return false;
            }
            initialMatchSize[i] = initialMatchesAll[i].size();
        }
//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1)
//cout <<  "initialMatchSize: " << initialMatchSize[i] << endl;
    }

    /// Find embeddings. Similar to GraMi + DFS Tree approach.
    std::vector<int> updatedMatchSize = initialMatchSize;
    std::vector<std::unordered_set<int>> nodeRepository(queryGraphInfo.nNodes); // Maintains distinct node embeddings for each query node.
    bool emptyBin = true;
    int thisBin = -1;
    int requiredEmbs = 0;


if(emptyBin){
            int seedVertex;
            if(!initialMatchesAll[queryOrderPerms[i][0]].empty()){ // if initial vertex stack is not yet empty |
                seedVertex = initialMatchesAll[queryOrderPerms[i][0]].back();
                initialMatchesAll[queryOrderPerms[i][0]].pop_back();
            }
            else
                continue;

        if(subGraph.FindMatchesSingleMNI(freq, seedVertex, queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, updatedMatchSize[queryOrderPerms[i][0]], matchedSubgraph.invalidCands, nodeRepository, requiredEmbs))
            return true;


        for(size_t b = 0; b < queryGraphInfo.nNodes; ++b){
            if(nodeRepository[b].size() > freq){
                thisBin = queryOrderPerms[b][0];
                emptyBin = false;
            }
        }

//cout << "updatedMatchSize[queryOrderPerms[i][0]] " << updatedMatchSize[queryOrderPerms[i][0]] << endl;
        if(updatedMatchSize[queryOrderPerms[i][0]] < freq)
            return false; // initial match size shrunk below the support value |

}


    bool stackNotEmpty = true;
int j = 0, requiredEmbs = 0;
    while(stackNotEmpty){
        std::set<int> supportReached;
        for(size_t i = 0; i < nodeRepository.size(); ++i)
            if (nodeRepository[i].size() >= freq)
                supportReached.insert(i);


//if(queryGraphInfo.nNodes == 3 && *matchedSubgraph.edgeLabels.rbegin()->second.begin() == 2 && *matchedSubgraph.edgeLabels.begin()->second.begin() == 1)
//cout << "supportReached:                              " << supportReached.size() << endl;
        int stackMatchCount = 0;
        for(size_t i = 0; i < queryOrderPerms.size(); ++i){
            if(!supportReached.empty())
                if(supportReached.find(*queryOrderPerms[i].rbegin()) == supportReached.end()){
                    ++stackMatchCount;
                    continue;
                }

            int seedVertex;
            if(!initialMatchesAll[queryOrderPerms[i][0]].empty()){ // if initial vertex stack is not yet empty |
                seedVertex = initialMatchesAll[queryOrderPerms[i][0]].back();
                initialMatchesAll[queryOrderPerms[i][0]].pop_back();
            }
            else{
                ++stackMatchCount;
                continue;
            }
            bool matchSolutions = false;
            while(!matchSolutions){
                if(nodeRepository[queryOrderPerms[i][0]].find(seedVertex) == nodeRepository[queryOrderPerms[i][0]].end()){ // if the next initial seed vertex is never matched before |
                    matchSolutions = true; // an attempt is being made to find/match solutions |
                    if(subGraph.FindMatchesSingleMNI(freq, seedVertex, queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, updatedMatchSize[queryOrderPerms[i][0]], matchedSubgraph.invalidCands, nodeRepository, requiredEmbs)){
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
cout << "# Embeddings    (Success): " << requiredEmbs << endl;
                        return true;
                    }
                    else{

                        if(updatedMatchSize[queryOrderPerms[i][0]] < freq)
                            return false; // initial match size shrunk below the support value |
                    }
                }
                else{
                    if(!initialMatchesAll[queryOrderPerms[i][0]].empty()){ // if initial vertex stack is not yet empty |
                        seedVertex = initialMatchesAll[queryOrderPerms[i][0]].back();
                        initialMatchesAll[queryOrderPerms[i][0]].pop_back();
                    }
                    else{
                        ++stackMatchCount;
                        break;
                    }
                }
            }
        }

        if(stackMatchCount == queryOrderPerms.size()) // each of the stack bin has gone empty.
            stackNotEmpty = false;
        ++j;
    }

    for(size_t i = 0; i < queryGraphInfo.nNodes; ++i){
        if(!nodeRepository[i].empty())
            initialMatchesAll[queryOrderPerms[i][0]] = FindDifferenceSet(initialMatchesAll[queryOrderPerms[i][0]], nodeRepository[i]);
//        if(thisBin == queryOrderPerms[i][0]){
//            continue;
//        }
//clock_t Time = clock();
        if(subGraph.FindMatchesFastMNI(freq, initialMatchesAll[queryOrderPerms[i][0]], queryGraphInfo, queryOrderPerms[i], nodeMatches, indexType.neighborTrie, matchedSubgraph.invalidCands, nodeRepository, requiredEmbs)){
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
cout << "# Embeddings    (Success): " << requiredEmbs << endl;
            return true;
        }
//        else{
//matchedSubgraph.timeEval.EmbEv += double(clock()-Time)/CLOCKS_PER_SEC;
//        }
    }
cout << "# Embeddings (Failure): " << requiredEmbs << endl;
    return false;

}
*/


/*
IsIsomorphicCurrent
IsAlreadyFSG
IsIsomorphicPrevious
IsPreviousNonFrequent
    getInvalidCands
    IsFrequent

IsAlreadyFSG
IsIsomorphicPrevious
IsPreviousNonFrequent
    getInvalidCands
    IsFrequent
*/
