#ifndef INDEX_H
#define INDEX_H

#include "../include/File.h"
#include "../include/rtree/RStarTree.h"
#include "../sumgra/Trie.h"

#include <stack>
#include <unordered_set>
#include <utility>

//#define DIM 700 // Expected maximum number of dimensions in the data set
#define SYN_SIZE 6 // Number of fields in the Synopses vector


//typedef std::map< std::pair<int, int>, std::set<int>> EdgeLabelMap;
//typedef std::map< std::pair<int, int>, std::bitset<DIM>> BitLabelMap;
//typedef std::vector<std::vector<std::bitset<DIM>>> EdgeLabelBit;

typedef RStarTree<int, SYN_SIZE, 32, 64> RTree;
typedef RTree::BoundingBox BoundingBox;

BoundingBox bounds(std::vector<short> synopses);

struct Visitor {
	int cnt;
	bool ContinueVisiting;
	std::vector<int> edgeIndices;
	Visitor() : cnt(0), ContinueVisiting(true) {};

	void operator()(const RTree::Leaf * const leaf)
	{
		edgeIndices.push_back(leaf->leaf);
	}
};

class Index
{
    public:
        Index();
        virtual ~Index();
        void BuildAttHash(const VecOfSet& attSign, AttMap& attributeHash);
        void QueryAttHash(const VecOfSet& queryAtt, const AttMap& attributeHash, VecOfSet&  attMatches);
        void CreateSynopses(const std::vector<std::set<int>>& signature, std::vector<short>& synopses);
        void SortSignature(EdgeLabel& neighbourSign, std::vector<int>& sortedNodes, const int& elements);
        void BuildSynTrie(const EdgeLabel& dataSignature, const int& dataNodes, RTree& synopsesTrie);
        void QuerySynTrie(const std::vector<std::set<int>>& initSignature, RTree& synopsesTrie, std::vector<int>& initialMatches);
        void BuildNeighTrie(const Vector2D& adjacencyList, const EdgeLabelMap& eLabelMap, const int& dataNodes, std::vector<Trie*>& nbrIndex);
        void QueryNeighTrie(Trie* t, const std::set<int>& multi_e, std::vector<int>& MatchedIds);
        void BuildBitSign(const EdgeLabel& dataSignature, const int& dataNodes, EdgeLabelBit& dataBitSet);
    protected:
    private:
        BoundingBox bounds(std::vector<short> synopses);
};

#endif // INDEX_H
