#ifndef MATCH_H
#define MATCH_H

#include "../include/File.h"
#include "Index.h"


struct Flags
{
    bool embFound;
    bool satBinFilled = false;
    bool validRule;
    bool invalidMatch;
    bool temp = false; // needed for some tests.
};

struct StarSolutions
{
    Vector coreSolutions;
    Vector3D satSolutions;
};

struct InitialVertex
{
    int pattern;
    int data;
};

struct Solutions
{
    Vector2D coreSolutions; // each row is a core vertex; if coreSolutions.size() > 1 => there are more than one core vertices with satellite solutoins |
    Vector3D sattSolutions; //
};

class Match
{
    public:
        Match();
        virtual ~Match();
        void OrderVertices(const int& queryNodes, const EdgeLabel& queryNeighbourSign, const Vector2D& queryAdjacencyList, std::vector<int>& querySequence);
        void OrderVerticesMNI(const int& queryNodes, const EdgeLabel& queryNeighbourSign, const Vector2D& queryAdjacencyList, Subgraph& matchedSubgraph);
        void OrderVerticesAllMNI(const int& queryNodes, const EdgeLabel& queryNeighbourSign, const Vector2D& queryAdjacencyList, std::vector<int>& orderedNodes, int& initialVertex);
        void OrderStarNodes(const EdgeLabel& neighbourSign, std::vector<int>& orderedQuery);
        void OrderInitStarNodes(const std::map<int, std::set<int>>& satNodes, Vector& orderedQueryInit);
        void OrderSatelliteNodes(GraphParameter& queryGraphInfo, std::map<int, std::set<int>>& satNodes, std::map<int, std::vector<int>>& orderedSatNodes);
        void OrderComplexNodes(GraphParameter& queryGraphInfo, const std::map<std::set<int>, int>&  edgeFrequency, std::map<int, std::set<int>>& satNodes, GraphParameter& truncPattern, std::map<int, std::vector<int>>& orderedSatellite, Vector& orderedCore);

        void GetSatelliteInfo(const Vector2D& adjacencyList, std::map<int, std::set<int>>& satNodes, std::set<std::pair<int,int>>& satEdges);
        void DecomposePattern(const std::set<std::pair<int,int>>& satEdges, GraphParameter& queryGraphInfo, GraphParameter& truncPattern);


        void FindMatches(const std::vector<int>& initialMatches, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie, Vector2D& embeddings);
        bool FindMisFrequentMatches(const int& freq, const std::vector<int>& initialMatches, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie);
        bool FindMniFrequentMatches(const int& freq, const std::vector<int>& initialMatches, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie);
        bool FindMatchesMNI(const int& freq, const std::vector<int>& initialMatches, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie, std::map<int, std::vector<int>>& invalidCand);
        bool FindMatchesSingleMNI(const int& freq, const int& initialMatch, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie, int& updatedMatchSize, std::map<int, std::vector<int>>& invalidCands, std::vector<std::unordered_set<int>>&  nodeRepository, int& requiredEmbs);
        bool FindMatchesFastMNI(const int& freq, const std::vector<int>& initialMatches, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie, std::map<int, std::vector<int>>& invalidCands, std::vector<std::unordered_set<int>>&  nodeRepository, bool& notFrequent);
        bool FindStarMatches(const int& freq, const Vector& initialMatches, GraphParameter& queryGraphInfo, const Vector& orderedQuery, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie, InvalidCands& invalidCands, NodeRepository&  nodeRepository);
        bool FindComplexMatches(const int& freq, const Vector& initialMatches, GraphParameter&  queryGraphInfo, GraphParameter& truncPattern, const std::map<int, std::set<int>>& satNodes, const std::map<int, std::vector<int>>& orderedSatNodes, const Vector& orderedCore, const Vector& orderedCoreOld, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie, InvalidCands& invalidCands, NodeRepository&  nodeRepository, bool& notFrequent);
        bool FindSatelliteMatches(const InitialVertex& initVertex, GraphParameter& queryGraphInfo, const Vector& orderedQuery, const std::vector<Trie*>& dataNeighTrie, Vector2D& patternMatches);
        bool CheckSI(std::vector<int>& initialMatches, const GraphParameter& queryGraphInfo, const std::vector<int>& querySequence, std::vector<int>& matchedDataVertices, const VecOfSet& nodeMatches, const std::vector<Trie*>& dataNeighTrie);
    protected:
    private:
        void chooseFrontier(const std::vector<int>& already_m, const Vector2D& queryAdjacencyList, std::vector<int>& frontier);
};

#endif // MATCH_H
